import numpy as np
import scipy.signal as sig

from .ITS_delay import ITS_delay_est


def sliding_delay_estimates(test, ref, fs, winLength=4, winStep=2):
    # SLIDING_DELAY_ESTIMATES perform sequence of windowed time delay estimates
    #
    #   SLIDING_DELAY_ESTIMATES(test,ref,fs) performs delay estimates between
    #       ref (input to the system under test) and test (output from system
    #       under test)
    #
    #   SLIDING_DELAY_ESTIMATES(test,ref,fs,winLength) specifies the window
    #       length in seconds, winLength, instead of using the default of 4
    #       seconds. The window length is the length of time to use for the
    #       delay estimates
    #
    #   SLIDING_DELAY_ESTIMATES(test,ref,fs,winLength,winStep) specifies the
    #       window step in seconds, winStep, instead of using the default of 2
    #       seconds. The window step is the amount of time the window moves
    #       forward for each subsequent time delay estimate.
    #
    #   Delays=SLIDING_DELAY_ESTIMATES(__) returns the delays as a vector
    #   instead of plotting them. Delays holds one delay estimate in ms for
    #   each time window. The resolution of these estimates is 1/8kHz. The
    #   accuracy of these estimates depends of the level of background noise
    #   and speech distortion in the input waveform
    #
    # Notes:  requires access to ITS_delay_est.m, written at ITS
    #   requires access to resample function included in Matlab Signal Processing Toolbox
    #
    # Written by Stephen Voran at the Institute for Telecommunication Sciences,
    # 325 Broadway, Boulder, Colorado, USA, svoran@its.bldrdoc.gov
    # March 30, 2016
    # Modification By Jesse Frey June, 2017

    # --------------------------Legal--------------------------
    # THE NATIONAL TELECOMMUNICATIONS AND INFORMATION ADMINISTRATION,
    # INSTITUTE FOR TELECOMMUNICATION SCIENCES ("NTIA/ITS") DOES NOT MAKE
    # ANY WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
    # WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR
    # A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.  THIS SOFTWARE
    # IS PROVIDED "AS IS."  NTIA/ITS does not warrant or make any
    # representations regarding the use of the software or the results thereof,
    # including but not limited to the correctness, accuracy, reliability or
    # usefulness of the software or the results.
    #
    # You can use, copy, modify, and redistribute the NTIA/ITS developed
    # software upon your acceptance of these terms and conditions and upon
    # your express agreement to provide appropriate acknowledgments of
    # NTIA's ownership of and development of the software by keeping this
    # exact text present in any copied or derivative works.
    #
    # The user of this Software ("Collaborator") agrees to hold the U.S.
    # Government harmless and indemnifies the U.S. Government for all
    # liabilities, demands, damages, expenses, and losses arising out of
    # the use by the Collaborator, or any party acting on its behalf, of
    # NTIA/ITS' Software, or out of any use, sale, or other disposition by
    # the Collaborator, or others acting on its behalf, of products made
    # by the use of NTIA/ITS' Software.

    # add test signal parameter
    test = np.array(test, dtype=np.float64)
    if np.any(np.isinf(test)):
        raise ValueError("Error with in test must be finite")
    # add reference signal parameter
    ref = np.array(ref, dtype=np.float64)
    if np.any(np.isinf(ref)):
        raise ValueError("Error values in ref must be finite")
    # add sample rate parameter
    if np.isinf(fs):
        raise ValueError("Error fs can not be infinite")
    # add window length argument
    if winLength <= 0:
        raise ValueError("Error winLength can not be less than zero")
    # add window separation argument
    if winLength <= 0:
        raise ValueError("Error winStep can not be less than zero")

    # sample rate to resample to
    # this is the rate that ITS_delay_est expects inputs to be in
    fs_re = 8e3

    # calculate resample factor
    n_re = fs / fs_re

    # check that resample rate is an integer
    if round(n_re) != n_re:
        # give error for invalid sample rate
        raise ValueError("fs must be an integer multiple of %i" % fs_re)

    # number of samples available in both files
    N = min(len(test), len(ref))
    # number of samples needed for each window
    Nwin = round(winLength * fs)
    # number of sampels to advance between windows
    Nstep = round(winStep * fs)

    firstSmp = 0
    lastSmp = firstSmp + Nwin - 1

    Delays = np.empty(0)  # Will hold delay estimates
    Times = np.empty(0)  # Will hold time of center of window

    # -----Loop to perform all possible time delay estimates-----
    while lastSmp < N:
        # get a section of the SUT input signal
        tempRef = ref[firstSmp : lastSmp + 1]
        # get a section of the SUT output signal
        tempTest = test[firstSmp : lastSmp + 1]
        # -----Apply our delay estimation tool to extracted portions of signal-----
        temp = ITS_delay_est(
            sig.resample_poly(tempRef, 1, n_re), sig.resample_poly(tempTest, 1, n_re), "f"
        )

        # -----Store results-----
        Delays = np.append(Delays, temp[1])
        Times = np.append(Times, 1 + (firstSmp + lastSmp) / 2)

        # -----Move window location ahead-----
        firstSmp = firstSmp + Nstep
        lastSmp = lastSmp + Nstep

    # convert from samples to ms
    Delays = Delays / 8
    # convert from samples to seconds
    Times = Times / fs

    # check if output arguments were given

    return Delays, Times



#!/usr/bin/env python

import csv
import glob
import json
import math
import mcvqoe.base
import os
import re
import shutil

import numpy as np


from .audio_chans import timecode_chans
from datetime import datetime, timedelta
from mcvqoe.base.terminal_user import terminal_progress_update
from .timecode import time_decode
from ..utilities.reprocess import get_module, reprocess_file


def test_name_parts(name):
    m=re.match( r'(?P<prefix>.*capture)_'+
                    '(?P<testtype>.+)_'+
                    '(?P<date>\d{2}-\w{3}-\d{4}_\d{2}-\d{2}-\d{2})'+
                    '(?P<ext>\.\w+)?$',
                name
              )
    if not m:
        raise RuntimeError(f'Unable to find test name parts from \'{name}\'')
    return (m.group('prefix'),m.group('testtype'),m.group('date'))

def timedelta_total_seconds(time):
    try:
        #try it as if it's an array
        return [timedelta_total_seconds(t) for t in time]
    except TypeError:
        #not an array, must be scalar time
        return time.days*(24*60*60) + time.seconds + time.microseconds*1e-6

#function to quickly find the index of the nearest value
#from https://stackoverflow.com/a/26026189
def find_nearest(array,value):
    idx = np.searchsorted(array, value, side="left")
    if idx > 0 and (idx == len(array) or math.fabs(value - array[idx-1]) < math.fabs(value - array[idx])):
        return idx-1
    else:
        return idx


def twoloc_process(tx_name, extra_play=0, rx_name = None, outdir="",
                        progress_update=terminal_progress_update,
                        align_mode='fit',
                        **kwargs       #get kwargs to accept arbitrary arguments
                   ):
    '''
    Process rx and tx files for a two location test.
    
    This writes a .csv file to data/csv and wave files to data/wav for a test. 

    Parameters
    ----------
    tx_name : string
        path to the transmit .csv file. If this is a relative path than
        `[outdir]/data/csv` is searched.
    extra_play : float, default=0
        Extra audio to add after tx clip stopped. This mayb be used, in some
        cases, to correct for data that was recorded with a poorly chosen
        overplay.
    rx_name : string, None
        Name of the receive .wav file. If this is a relative path than
        `[outdir]/data/csv` is searched. If this is None then `data/2loc_rx-data`
        is searched.
    outdir : string, default=""
        Directory that contains the `data/` folder where data will be read from
        and written to.
    progress_update : function, default=terminal_user
        Function to call with updates on processing progress. 
        
    See Also
    --------
        mcvqoe.mouth2ear : mouth to ear code, can produce 2 location data.
    '''

    #This software was developed by employees of the National Institute of
    #Standards and Technology (NIST), an agency of the Federal Government.
    #Pursuant to title 17 United States Code Section 105, works of NIST
    #employees are not subject to copyright protection in the United States and
    #are considered to be in the public domain. Permission to freely use, copy,
    #modify, and distribute this software and its documentation without fee is
    #hereby granted, provided that this notice and disclaimer of warranty
    #appears in all copies.
    #
    #THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
    #EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY
    #WARRANTY THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED
    #WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
    #FREEDOM FROM INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL
    #CONFORM TO THE SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR
    #FREE. IN NO EVENT SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT
    #LIMITED TO, DIRECT, INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING
    #OUT OF, RESULTING FROM, OR IN ANY WAY CONNECTED WITH THIS SOFTWARE,
    #WHETHER OR NOT BASED UPON WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER
    #OR NOT INJURY WAS SUSTAINED BY PERSONS OR PROPERTY OR OTHERWISE, AND
    #WHETHER OR NOT LOSS WAS SUSTAINED FROM, OR AROSE OUT OF THE RESULTS OF, OR
    #USE OF, THE SOFTWARE OR SERVICES PROVIDED HEREUNDER.
   
    #-----------------------------Validate inputs-----------------------------

    #normalize path
    tx_name = os.path.normpath(tx_name)

    if os.path.isdir(tx_name):
        #given directory, assume this is the .wav dir
        tx_wav_path = tx_name
        #find .csv files in the directory
        csvs = glob.glob(os.path.join(tx_wav_path, '*.csv'))

        if not csvs:
            raise RuntimeError(f'No .csv files found in \'{tx_wav_path}\'')
        elif len(csvs) > 1:
            raise RuntimeError(f'More than one .csv file found in \'{tx_wav_path}\'')

        tx_name = csvs[0]
    else:
        #get folder path from filename
        tx_wav_path=os.path.dirname(tx_name)

    #get folder name (just name, no path)
    tx_wav_fold =  os.path.basename(tx_wav_path)

    if not tx_wav_fold:
        raise RuntimeError(f'unable to extract wav folder from \'{tx_name}\'')

    #extract parts of tx name for validation
    (tx_prefix,tx_tt,tx_date)=test_name_parts(tx_wav_fold)
    #check prefix
    if(tx_prefix != 'Tx_capture'):
        raise ValueError(f'Unexpected filename prefix \'{tx_prefix}\' for tx file')
    
    #extra_play must be non-neg
    if extra_play < 0:
        raise ValueError("extra_play must be non negative")

    #tolerance for timecode variation
    tc_warn_tol = 0.0001

    # --------------------------[Locate input data]--------------------------
    
    #go two levels up from .csv file
    tx_fold=os.path.dirname(os.path.dirname(tx_name))
    
    indir=os.path.abspath(os.path.join(tx_fold,'..','..'))
    
    #check if rx_name is a directory
    if rx_name and os.path.isdir(rx_name):
        #use rx_name as dir
        rx_dir = rx_name
        #we don't have a specific name, clear rx_name
        rx_name = None
    else:
        rx_dir=os.path.join(indir,'data','2loc_rx-data')

    # -----------------------[Setup Files and folders]-----------------------

    # generate data dir names
    data_dir = os.path.join(outdir, "data")
    wav_data_dir = os.path.join(data_dir, "wav")
    csv_data_dir = os.path.join(data_dir, "csv")

    # create data directories
    os.makedirs(csv_data_dir, exist_ok=True)
    os.makedirs(wav_data_dir, exist_ok=True)

    # generate base file name to use for all files
    base_filename ='_'.join(('capture2',tx_tt,tx_date))

    # generate test dir names
    wavdir = os.path.join(wav_data_dir, base_filename)

    # create test dir
    os.makedirs(wavdir, exist_ok=True)

    # generate csv name
    csv_out_name = os.path.join(csv_data_dir, f"{base_filename}.csv")
    
    #------------------Find appropriate rx file by timecode------------------
    #if a rx file name is not specified, find the appropriate file in rx-dat 
    #folder. if there are more than one suitable rx files based on timecode, 
    #this will find the one with the smallest delay, with delay defined as 
    #difference between the start times of the rx and tx recordings
    
    #if no file was specified for the rx file, search for it in rx-dat
    if not rx_name:
    
        #attempt to get date from tx filename
        tx_date=datetime.strptime(tx_date, '%d-%b-%Y_%H-%M-%S')
        
        #rx_files is a dict with the delays as keys, and the rx path as values
        rx_files = {}
        progress_update('status', 0, 0, msg=f'looking for rx files in \'{rx_dir}\'')
        #loop thru all rx files
        for rx_file_name in glob.glob(os.path.join(rx_dir,'*.wav')):
            progress_update('status', 0, 0, msg=f'Looking at {rx_file_name}')
            #strip leading folders
            rx_basename=os.path.basename(rx_file_name)
            #split into parts
            (rx_prefix,rx_tt,rx_date)=test_name_parts(rx_basename)
            #validate that this is a correct rx file
            if rx_prefix != 'Rx_capture':
                #give error
                progress_update('warning', 0, 0, msg=f'Rx filename "{rx_basename}" is not in the proper form. Can not determine Rx filename')
                #if not a correct rx file, skip this file and go to next one
                continue
            rx_start_date=datetime.strptime(rx_date, '%d-%b-%Y_%H-%M-%S')
            #add to the rx_file dict, with delays as the key, and full path as value
            rx_files[tx_date - rx_start_date] = rx_file_name
    
        #create a np array of all of delays
        delays = np.array(list(rx_files), dtype=timedelta)
        
        #find the smallest positive delay
        minDelay = min(delays[delays > timedelta()])
        
        #find the file with the the smallest delay
        rx_name=rx_files[minDelay]

        progress_update('status', 0, 0, msg=f'Loading {rx_name}')

        #read file
        rx_fs, rx_dat = mcvqoe.base.audio_read(rx_name)
    
        #find the duration of the rx file
        duration = timedelta(seconds=len(rx_dat)/rx_fs)
        
        #check that tx date falls within rx file time
        if minDelay < duration:
            rx_name = rx_files[minDelay]
        #otherwise there is no suitable rx file
        else:
            raise ValueError("Could not find suitable Rx file")
    
    else:
        rx_fs, rx_dat = mcvqoe.base.audio_read(rx_name)
    
    rx_dat=mcvqoe.base.audio_float(rx_dat)
    
    #--------------------Prep work for calculating delays--------------------

    rx_info_name=os.path.splitext(rx_name)[0]+'.json'
    
    with open(rx_info_name) as info_f:
        rx_info=json.load(info_f)
        
    tc_chans=timecode_chans(rx_info['channels'])
    if not tc_chans:
        raise ValueError(f'Timecode channel could not be found in {rx_info["channels"]}')
    
    #use the first one
    rx_tc_idx=tc_chans[0]
    
    #timecode type
    rx_tc_type=rx_info['channels'][rx_tc_idx]
    
    #get channels
    rx_extra_chans=rx_info['channels']
    #remove timecode channel
    del rx_extra_chans[rx_tc_idx]
    
    #decode the rx timecode
    rx_time, rx_snum = time_decode(rx_tc_type,rx_dat[:,rx_tc_idx], rx_fs)
    
    #make rx_time a numpy array
    rx_time = np.array(rx_time)

    if align_mode == 'interpolate':
        #we are interpolating, get reference time
        ref_time =  rx_time[0]
        #interpolate so we have intermediate values
        rx_interp = np.interp(range(len(rx_dat)),rx_snum,timedelta_total_seconds(rx_time-ref_time))
    elif align_mode == 'fit':
        #we are fitting, get reference time
        ref_time =  rx_time[0]
        #fit index vs time
        #do a linear fit of the timecode data to get time vs index
        rx_fit = np.polyfit(timedelta_total_seconds(rx_time-ref_time), rx_snum, 1)
        #get model
        rx_idx_fun = np.poly1d(rx_fit)

    extra_samples = extra_play * rx_fs
    with open(tx_name,'rt') as tx_csv_f, open(csv_out_name,'wt',newline='') as out_csv_f:
        
        #create dict reader
        reader=csv.DictReader(tx_csv_f)
        
        #create dict writer, same fields as input
        writer=csv.DictWriter(out_csv_f,reader.fieldnames)
        
        #write output header
        writer.writeheader()
        
        #get data from file
        #NOTE : this may not work well for large files! but should, typically, be fine
        rows = tuple(reader)
        
        #get total trials for progress
        total_trials = len(rows)
    
        #loop through all tx recordings
        for trial,row in enumerate(rows):
            
            progress_update('proc', total_trials, trial)
            
            tx_rec_name = f'Rx{trial+1}_{row["Filename"]}.wav'
            full_tx_rec_name = os.path.join(tx_wav_path, tx_rec_name)

            #check if file exists
            if not os.path.exists(clip_path):
                #update progress
                progress_update('status', total_trials, trial,
                msg = 'Attempting to decompress audio...')
                #unzip audio if it exists
                mcvqoe.base.Measure.unzip_audio(audio_path)

            tx_rec_fs, tx_rec_dat = mcvqoe.base.audio_read(full_tx_rec_name)
            #make floating point for processing purposes
            tx_rec_dat=mcvqoe.base.audio_float(tx_rec_dat)
            
            tx_rec_chans=mcvqoe.base.parse_audio_channels(row['channels'])
            
            if(len(tx_rec_chans)==1):
                #only one channel, make sure that it's a timecode
                if('timecode' not in tx_rec_chans[0]):
                    raise ValueError(f'Expected timecode channel but got {row["channels"][0]}')
                
                #make sure that timecode types match
                if rx_tc_type != tx_rec_chans[0]:
                    raise ValueError(f'Tx timecode type is {tx_rec_chans[0]} but Rx timecode type is {rx_tc_type}')
                
                #one channel, only timecode
                tx_rec_tca=tx_rec_dat
                
                tx_extra_audio = None
                tx_extra_chans = None
            else:                
                #grab the same type of timecode we used for Rx
                tx_time_idx = tx_rec_chans.index(rx_tc_type)

                tx_rec_tca = tx_rec_dat[:,tx_time_idx]

                #extra channels
                tx_extra_audio = np.delete(tx_rec_dat,tx_time_idx,1)

                #copy to new array without timecode channel
                tx_extra_chans = list(tx_rec_chans)
                del tx_extra_chans[tx_time_idx]
            
            #decode timecode
            tx_time, tx_snum = time_decode(rx_tc_type, tx_rec_tca, tx_rec_fs)

            if align_mode == 'fixed':
                #array for matching sample numbers
                tx_match_samples = []
                rx_match_samples = []


                for time,snum in zip(tx_time,tx_snum):
                    
                    #calculate difference from rx timecode
                    time_diff = abs(rx_time - time)
                    
                    #find minimum difference
                    min_v = np.amin(time_diff)

                    #check that difference is small
                    if min_v < timedelta(seconds=0.5):

                        #get matching index
                        idx=np.argmin(time_diff)

                        #append sample number
                        tx_match_samples.append(snum)
                        rx_match_samples.append(rx_snum[idx])

                #get matching frame start indicies
                mfr=np.column_stack((tx_match_samples, rx_match_samples))

                #get difference between matching timecodes
                mfd=np.diff(mfr, axis=0)


                #get ratio of samples between matches
                mfdr = mfd[:,0] / mfd[:,1]

            
                if not np.all(np.logical_and(mfdr < (1+tc_warn_tol), mfdr>(1-tc_warn_tol))):
                    progress_update('warning', total_trials, trial, f'Timecodes out of tolerence for trial {trial+1}. {mfdr}')

                #calculate first rx sample to use
                first=mfr[0,1]-mfr[0,0]

                #calculate last rx sample to use
                last=mfr[-1,1]+len(tx_rec_tca)-mfr[-1,0]+extra_samples - 1
            elif align_mode == 'interpolate' or align_mode =='fit':
                tx_tnum =timedelta_total_seconds(tx_time - ref_time)

                #do a linear fit of the timecode data to get index vs time
                fit = np.polyfit(tx_snum, tx_tnum, 1)
                #get model
                tc_fun = np.poly1d(fit)

                #get time of start and end of tx clip
                tx_start_time = tc_fun(0)
                tx_end_time = tc_fun(len(tx_rec_tca) + extra_samples - 1)

                if align_mode == 'interpolate':
                    #get indices in the Rx array
                    first = find_nearest(rx_interp, tx_start_time)
                    last  = find_nearest(rx_interp, tx_end_time)
                elif align_mode == 'fit':
                    first = math.floor(rx_idx_fun(tx_start_time))
                    last  = math.ceil(rx_idx_fun(tx_end_time))

            else:
                raise ValueError(f'Invalid value, \'{align_mode}\' for align_mode')
            #get rx recording data from big array
            rx_rec=rx_dat[first:last+1,:]
            #remove timecode
            rx_rec = np.delete(rx_rec,rx_tc_idx,1)

            if tx_extra_chans:
                #add tx extra chans to rx extra chans
                out_chans=tuple(rx_extra_chans+tx_extra_chans)

                #get the length of the longest array
                new_len = np.max((rx_rec.shape[0],tx_extra_audio.shape[0]))

                #resize recording
                rec_shape = list(rx_rec.shape)
                rec_shape[0] = new_len
                rx_rec.resize( tuple(rec_shape))

                #resize tx extra
                tx_shape = list(tx_extra_audio.shape)
                tx_shape[0] = new_len
                tx_extra_audio.resize( tuple(tx_shape))

                #both arrays should now be the same length (in time)
                out_audio=np.column_stack((rx_rec,tx_extra_audio))
            else:
                #no extra chans, all out chans from rx
                out_chans=rx_extra_chans
                out_audio=rx_rec
            
            #overwrite new channels to csv
            row['channels']=mcvqoe.base.audio_channels_to_string(out_chans)
            
            #TODO : process audio?
            
            # Create audiofile name/path for recording
            audioname = f'Rx{trial+1}_{row["Filename"]}.wav'
            audioname = os.path.join(wavdir, audioname)
            
            #save out Rx recording as given data type
            mcvqoe.base.audio_write(audioname,rx_fs,out_audio)
            
            #write row to new .csv
            writer.writerow(row)
        
        #copy Tx files into destination folder
        for name in glob.glob(os.path.join(tx_wav_path,'Tx_*')):
            #get clip name from path
            clip_name=os.path.basename(name)
            #construct destination name
            destname=os.path.join(wavdir,clip_name)
            #copy file
            shutil.copyfile(name,destname)

    #return output filename
    return csv_out_name


def main():
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument("tx_name",
                        type=str,
                        help='Name of the Tx .csv file to process/'
                        )
    parser.add_argument("--extra-play",
                        type=int,
                        default=0,
                        help='Duration of extra audio to add after tx clip '+
                        'stopped. This mayb be used, in some cases, to correct '+
                        'for data that was recorded with a poorly chosen overplay.'
                        )
    parser.add_argument("--rx-name",
                        type=str,
                        default=None,
                        help='Filename of the rx file to use. If a directory '+
                        'is given, it will be searched for files'
                        )
    parser.add_argument("--outdir",
                        type=str,
                        default="",
                        help='Root of directory structure where data will be stored'
                        )
    parser.add_argument('-m', '--measurement',
                        type=str,
                        default=None,
                        metavar='M',
                        help='measurement to use to do reprocessing'
                        )

    args = parser.parse_args()

    #try to load measurement class
    try:
        measurement_class = get_module(module_name=args.measurement, datafile=args.tx_name)
    except (RuntimeError,KeyError) as e:
        terminal_progress_update('warning', 0, 0, msg=f'Unable to determine measurement. Output file will not be processed')
        #set measurement class for later
        measurement_class = None

    out_name = twoloc_process(**vars(args))

    if measurement_class:
        #create test obj to reprocess with
        test_obj=measurement_class()

        reprocess_file(test_obj, out_name, out_name)

        print(f'Reprocessing complete for \'{out_name}\'')

if __name__ == '__main__':
    main()


import datetime
import os
import sys
import tkinter as tk
from tkinter import scrolledtext

import pkg_resources


class TestInfoGui(tk.Tk):
    """
    Class to show a gui to get information on a test.

    This class shows a GUI where the user can input parameters for a test. The
    test info defaults are logged to a file so that, when running multiple
    similar tests, some time can be saved.

    Attributes
    ----------
    defaults_file : str, default="test-type.txt"
        the file where defaults will be read/stored
    chk_audio_function : function, default=None
        The function to call when the check audio button is pressed. If this is
        None, no check audo button is used.
    outdir : str, default=''
        directory to write defaults_file in
    write_test_info : bool
        if true defaults will be written to the `defaults_file` after submit is
        pressed in the GUI. otherwise no notes will be written.
    info_in : dict
        Dictionary of info that will be used for GUI defaults
    test_info : dict
        Dictionary of info from the user, containing test info.

    See Also
    --------
    post_test : Connivance function for TestInfoGui.
    PostTestGui : GUI for post test notes.

    Examples
    --------
    get test notes from the user

    >>> gui=TestInfoGui()
    >>> test_info=gui.show()
    >>> print(test_info)
    """

    def __init__(
        self,
        defaults_file="test-type.txt",
        chk_audio_function=None,
        outdir="",
        write_test_info=True,
        *args,
        **kwargs,
    ):
        tk.Tk.__init__(self, *args, **kwargs)
        """
        Class to show a gui to get information on a test.

        Parameters
        ----------
        defaults_file : str, default="test-type.txt"
            the file where defaults will be read/stored
        chk_audio_function : function, default=None
            The function to call when the check audio button is pressed. If this is
            None, no check audo button is used.
        outdir : str, default=''
            directory to write defaults_file in
        write_test_info : bool, default=True
            if true defaults will be written to the `defaults_file` after submit is
            pressed in the GUI. otherwise no notes will be written.
        """

        self.defaults_name = os.path.join(outdir, defaults_file)
        self.load_settings(self.defaults_name)
        self.chk_audio_function = chk_audio_function
        self.write_test_info = write_test_info

        # Show MCV logo
        img_file = pkg_resources.resource_filename("mcvqoe.gui", "MCV-sm.png")
        self.tk.call("wm", "iconphoto", self._w, tk.PhotoImage(file=img_file))

    def load_settings(self, fname):
        """
        Load test info defaults from a file.

        Parameters
        ----------
        fname : str
            Full path of the file to read.
        """
        self.info_in = {}

        if fname:
            # --------------------[Obtain Previous Test info]--------------------
            try:
                with open(fname, "r") as prev_test:
                    self.info_in["test_type"] = prev_test.readline().split('"')[1]
                    self.info_in["tx_dev"] = prev_test.readline().split('"')[1]
                    self.info_in["rx_dev"] = prev_test.readline().split('"')[1]
                    self.info_in["system"] = prev_test.readline().split('"')[1]
                    self.info_in["test_loc"] = prev_test.readline().split('"')[1]
            except FileNotFoundError:
                self.info_in["test_type"] = ""
                self.info_in["tx_dev"] = ""
                self.info_in["rx_dev"] = ""
                self.info_in["system"] = ""
                self.info_in["test_loc"] = ""
            # dummy value for pre test notes, not saved in file
            self.info_in["Pre Test Notes"] = ""

    def show(self):
        """
        Populate window and wait for user input.

        Returns
        -------
        dict
            Dictionary with "Post Test Notes" or "Error Notes".
        """
        # Window creation
        self.title("Test Information")

        # End the program if the window is exited out
        self.protocol("WM_DELETE_WINDOW", self._cancel_action)

        # Test type prompt
        l1 = tk.Label(self, text="Test Type")
        l1.grid(row=0, column=0, padx=10, pady=5, sticky=tk.W)

        self.test_type_edit = tk.Entry(self, bd=2, width=50)
        self.test_type_edit.insert(tk.END, "")
        self.test_type_edit.insert(0, self.info_in["test_type"])
        self.test_type_edit.grid(row=1, column=0, padx=10, pady=5, sticky=tk.W)
        self.test_type_edit.focus()

        # Transmit device prompt
        l2 = tk.Label(self, text="Transmit Device")
        l2.grid(row=2, column=0, padx=10, pady=5, sticky=tk.W)
        self.tx_dev_edit = tk.Entry(self, bd=2)
        self.tx_dev_edit.insert(0, self.info_in["tx_dev"])
        self.tx_dev_edit.grid(row=3, column=0, padx=10, pady=5, sticky=tk.W)

        # Receive device prompt
        l3 = tk.Label(self, text="Receive Device")
        l3.grid(row=4, column=0, padx=10, pady=5, sticky=tk.W)
        self.rx_dev_edit = tk.Entry(self, bd=2)
        self.rx_dev_edit.insert(0, self.info_in["rx_dev"])
        self.rx_dev_edit.grid(row=5, column=0, padx=10, pady=5, sticky=tk.W)

        # System prompt
        l4 = tk.Label(self, text="System")
        l4.grid(row=6, column=0, padx=10, pady=5, sticky=tk.W)
        self.system_edit = tk.Entry(self, bd=2, width=60)
        self.system_edit.insert(0, self.info_in["system"])
        self.system_edit.grid(row=7, column=0, padx=10, pady=5, sticky=tk.W)

        # Test location prompt
        l5 = tk.Label(self, text="Test Location")
        l5.grid(row=8, column=0, padx=10, pady=5, sticky=tk.W)

        self.test_loc_edit = tk.Entry(self, bd=2, width=100)
        self.test_loc_edit.insert(0, self.info_in["test_loc"])
        self.test_loc_edit.grid(row=9, column=0, padx=10, pady=5, sticky=tk.W)

        # Pre-test notes prompt
        l6 = tk.Label(self, text="Please enter notes on pre-test conditions")
        l6.grid(row=10, column=0, padx=10, pady=5, sticky=tk.W)
        self.pre_notes_edit = scrolledtext.ScrolledText(self, bd=2, width=100, height=15)
        self.pre_notes_edit.insert(tk.INSERT, self.info_in["Pre Test Notes"])
        self.pre_notes_edit.grid(row=11, column=0, padx=10, pady=5, sticky=tk.W)

        # 'Submit', 'Test', and 'Cancel' buttons
        button_frame = tk.Frame(self)
        button_frame.grid(row=12, column=1, sticky=tk.E)

        exit_frame = tk.Frame(self)
        exit_frame.grid(row=12, column=0, sticky=tk.W)

        if self.chk_audio_function is not None:
            button = tk.Button(exit_frame, text="Check Audio", command=self.chk_audio_function)
            button.grid(row=0, column=0, padx=10, pady=10)

        button = tk.Button(button_frame, text="Submit", command=self._submit_action)
        button.grid(row=0, column=0, padx=10, pady=10)

        button = tk.Button(button_frame, text="Cancel", command=self._cancel_action)
        button.grid(row=0, column=1, padx=10, pady=10)

        # Run Tkinter window
        self.mainloop()

        if self.test_info is not None and self.defaults_name is not None and self.write_test_info:
            # --------------------[Print Test Type and Test Notes]----------------------

            # Print info to screen
            print("\nTest type: %s\n" % self.test_info["Test Type"], flush=True)
            print("Pre test notes:\n%s" % self.test_info["Pre Test Notes"], flush=True)

            # Write info to .txt file
            with open(self.defaults_name, "w") as file:
                file.write('Test Type : "%s"\n' % self.test_info["Test Type"])
                file.write('Tx Device : "%s"\n' % self.test_info["Tx Device"])
                file.write('Rx Device : "%s"\n' % self.test_info["Rx Device"])
                file.write('System    : "%s"\n' % self.test_info["System"])
                file.write('Test Loc  : "%s"\n' % self.test_info["Test Loc"])

        # can't seem to find documentation on this but, it fixes the following:
        # Tcl_AsyncDelete: async handler deleted by the wrong thread
        self.quit()

        return self.test_info

    def _submit_action(self):
        """
        Collect user input from Tkinter input window.
        """
        self.test_info = {
            "Test Type": self.test_type_edit.get(),
            "Tx Device": self.tx_dev_edit.get(),
            "Rx Device": self.rx_dev_edit.get(),
            "System": self.system_edit.get(),
            "Test Loc": self.test_loc_edit.get(),
            "Pre Test Notes": self.pre_notes_edit.get(1.0, tk.END),
        }
        # Delete window
        self.destroy()

    def _cancel_action(self):

        # signal cancel by setting test_info to None
        self.test_info = None

        self.destroy()


class PostTestGui(tk.Tk):
    """
    Class to show a gui to get notes after a test is complete.

    This class is used to show a GUI to the user on completion of a test to get
    any information about how the test went. If an error has been encountered
    during the test this is displayed in the gui and "Error Notes" are collected.
    Otherwise "Post Test Notes" are collected.

    Attributes
    ----------
    err : Exception
        If an exception is being handled, the exception, otherwise None.

    See Also
    --------
    post_test : Convenience function for PostTestGui.
    TestInfoGui : GUI to gather test info.

    Examples
    --------
    Get post test notes from user

    >>> gui=PostTestGui()
    >>> notes=gui.show()
    >>> print(notes)
    """

    def __init__(self, err, *args, **kwargs):
        """
        Class to show a gui to get notes after a test is complete

        Parameters
        ----------
        err :
            If an exception is being handled, the exception, otherwise None.
        """

        tk.Tk.__init__(self, *args, **kwargs)

        # error info, this will be used for the user prompt and to determine what
        # kind of notes we will create
        self.err = err

        # Show MCV logo
        img_file = pkg_resources.resource_filename("mcvqoe.gui", "MCV-sm.png")
        self.tk.call("wm", "iconphoto", self._w, tk.PhotoImage(file=img_file))

    def show(self):
        """
        Populate window and wait for user input.

        Returns
        -------
        dict
            Dictionary with "Post Test Notes" or "Error Notes".
        """

        # ------------------------------[Setup GUI]------------------------------

        self.title("Test Information")
        self.after(1, self.focus_force())

        # Prevent error if user exits
        self.protocol("WM_DELETE_WINDOW", self._collect_notes)

        # Pre-test notes prompt
        if self.err:
            label = tk.Label(
                self,
                text=f'An "{self.err.__name__}" was encountered. Please enter notes on test conditions',
            )
        else:
            label = tk.Label(self, text="Please enter post-test notes")
        label.grid(row=0, column=0, padx=10, pady=5, sticky=tk.W)

        self.notes_edit = scrolledtext.ScrolledText(self, bd=2, width=100, height=15)
        self.notes_edit.grid(row=1, column=0, padx=10, pady=5)
        self.notes_edit.focus()

        # 'Submit' and 'Cancel' buttons
        button_frame = tk.Frame(self)
        button_frame.grid(row=2, column=0, sticky=tk.E)

        button = tk.Button(button_frame, text="Submit", command=self._collect_notes)
        button.grid(row=0, column=0, padx=10, pady=10)

        button = tk.Button(button_frame, text="Cancel", command=self._cancel_action)
        button.grid(row=0, column=1, padx=10, pady=10)

        # Run Tkinter window
        self.mainloop()

        # can't seem to find documentation on this but, it fixes the following:
        # Tcl_AsyncDelete: async handler deleted by the wrong thread
        self.quit()

        if self.err:
            return {"Error Notes": self.notes}
        else:
            return {"Post Test Notes": self.notes}

    def _collect_notes(self):
        """Collect user's post-test notes."""

        self.notes = self.notes_edit.get(1.0, tk.END)

        # Delete window
        self.destroy()

    def _cancel_action(self):
        """
        Helper function for when cancel is pressed.
        """

        # set notes to empty
        self.notes = ""
        # destroy window
        self.destroy()


def post_test(error_only=False):
    """
    Convenience function for PostTestGui.

    This function creates a PostTestGui, calls the show method and returns the
    info dict.

    Parameters
    ----------
    error_only : bool, default=False
        If true will only show a dialog if an error occurred.

    Returns
    -------
    dict
        Dictionary with "Post Test Notes" or "Error Notes".
    """

    # get current error status, will be None if we are not handling an error
    error = sys.exc_info()[0]

    # check if there is no error and we should only show on error
    if (not error) and error_only:
        # nothing to do, bye!
        return {}

    # make a gui object
    gui = PostTestGui(error)

    # show things and return notes
    return gui.show()


def pretest(outdir="", check_function=None):
    """
    Convenience function for TestInfoGui.

    This function creates a TestInfoGui, calls the show method and returns test
    info. If no test info is given this function will call sys.exit.

    Parameters
    ----------
    outdir : str
        The directory to write test defaults to.
    check_function : function
        Function to call to check audio.

    Returns
    -------
    dict
        A dictionary of test info.
    """
    gui = TestInfoGui(chk_audio_function=check_function)

    test_info = gui.show()

    # check if the user canceled
    if test_info is None:
        print(f"\n\tExited by user")
        sys.exit(1)

    return test_info



from .resample_techniques import bootstrap_ci
from .resample_techniques import approx_permutation_test
from .resample_techniques import standard_error
from .resample_techniques import improved_autocorrelation
from .resample_techniques import bootstrap_datasets_ci


import csv

import numpy as np
import pandas as pd
import scipy.io.wavfile
import scipy.signal as sig


def audio_float(dat):
    """
    Convert audio data to float with standard scale.

    Data read with scipi.io.wavefile can be returned in a number of number
    formats. `audio_float` returns audio data in floating point for scaled to =/-1

    Parameters
    ----------
    dat : numpy array
        Audio data to convert to float.

    Returns
    -------
    numpy array
        Floating point audio data scaled to a range of +/-1.

    See Also
    --------
    scipi.io.wavefile : Functions for reading and writing .wav files.
    mcvqoe.audio_type : General function for converting audio to any valid type

    Examples
    --------
    load audio and ensure that it is on the -1 to +1 range

    >>> fs, dat = scipy.io.wavfile('audio.wav')
    >>> float_dat=audio_float(dat)
    """
    return audio_type(dat,dtype=np.dtype('float'))

def audio_type(dat,dtype = np.dtype('int16')):
    """
    Convert audio data to different data type with standard scale.

    Data read and written with scipi.io.wavefile can be in a variety of 
    different formats. `audio_type` returns audio data in different data types
    to ensure the proper format data is used for reading and writing.
    

    Parameters
    ----------
    dat : numpy array
        Audio data to convert.
    dtype : np.dtype, optional
        Output data type. The default is np.dtype('int16').

    Returns
    -------
    numpy array
        Array with audio scaled to appropriate values for new data type.
        
    See Also
    --------
    scipy.io.wavfile : Functions for reading and writing .wav files
    mcvqoe.audio_float : Convenience function for converting audio to float

    """
    
    # Return audio if it's already the right type
    if(dat.dtype == dtype):
        return(dat)
        
    #---------------------[Convert to float]------------------
    if np.issubdtype(dat.dtype, np.floating):
        dat = dat
    elif(dat.dtype is np.dtype('uint8')):
        dat = (dat.astype('float')-64)/64
    elif(dat.dtype is np.dtype('int16')):
        dat = dat.astype('float')/(2**15)
    elif(dat.dtype is np.dtype('int32')):
        dat = dat.astype('float')/(2**31)
    else:
        raise RuntimeError(f'unknown audio type \'{dat.dtype}\'')
        
    #-----------------[Convert to output]------------------------
    if np.issubdtype(dtype, np.floating):
        return(dat.astype(dtype))
    elif(dtype is np.dtype('uint8')):
        out_dat = 64*dat + 64
        return out_dat.astype('uint8')
    elif(dtype is np.dtype('int16')):
        out_dat = 2**15 * dat
        return out_dat.astype('int16')
    elif(dtype is np.dtype('int32')):
        out_dat = 2**31 * dat
        return out_dat.astype('int32')
    else:
        raise RuntimeError(f'unknown audio type \'{dat.dtype}\'')


def audio_read(filename):
    """
    Open a WAV file.

    Return the sample rate (in samples/sec) and data from a WAV file.
    Regardless of original WAV file format, data is returned as float32 numpy
    array.

    Parameters
    ----------
    fname : str
        Input WAV file.

    Returns
    -------
    sample_rate : int
        Sample rate of WAV file

    audio_data : numpy array
        Data read from WAV file. Data is 1-D for 1-channel WAV, or 2-D of
        shape (Nsamples, Nchannels) otherwise. Numpy data type is always set
        to float.

    See Also
    --------
    mcvqoe.base.audio_write

    mcvqoe.base.audio_type

    """
    sample_rate, audio_data = scipy.io.wavfile.read(filename)
    audio_data = audio_type(audio_data, dtype=np.dtype('float32'))
    return sample_rate, audio_data


def audio_write(filename, rate, data):
    """
    Write a NumPy array as a WAV file.

    Regardless of type of data, writes wavfile as 16 bit PCM.

    Parameters
    ----------
    filename : TYPE
        DESCRIPTION.
    rate : int
        The sample rate (in samples/sec).
    data : ndarray
        A 1-D or 2-D NumPy array of either integer or float data-type.

    Returns
    -------
    None.

    See Also
    --------
    mcvqoe.base.audio_read

    mcvqoe.base.audio_type

    """
    data = audio_type(data, dtype=np.dtype('int16'))
    scipy.io.wavfile.write(filename, rate, data)


def svp56_fast(x, fs=8000):
    """
    Digital Speech Voltmeter function for audio.

    This code implements the "Speech Voltmeter" given in CCITT Rec. P.56.
    It is a fast version of svp56.m which was written from the description in
    the hard copy doc that came with UGST STL 96.  svp56.m results have been
    verified against results of the C code from UGST.

    svp56_fast.m agrees with with svp56.m as follows.  It either agrees
    exactly (the integer counts in all 16 bits of "a" match exactly at
    inspection point A) or almost exactly (the integer counts in some of the
    16 bits of "a" are off by one at inspection point A, and the asl and saf
    differ in the 4th or 5th decimal place)
    Note that on a fixed platform svp56_fast.m requires about 8# of the
    running time that svp56.m requires

    Written by S. Voran, in mid 1990's

    Usage:  [asl,saf,active]=svp56_fast(x, fs)

    Parameters
    ----------
    x : list or str
        Speech signal or file name of a .wav file.
    fs : number, default=8000
        Sample rate of x.

    Returns
    -------
    asl : number
        Active speech level in dB (relative to 16 bits).
    saf : float
        Speech activity factor (0 to 1).
    active : tuple
        Same length as the speech signal. Contains 1's in active areas, 0's else.
        "active" is an approximation, and the average absolute error:
        mean(abs( saf - sum(active)/n )) is about 0.2# or 0.3#.

    See Also
    --------
    mcvqoe.a_weighted_power : Human ear model power level meter.

    Examples
    --------
    Measure the active speech level of a .wav file

    >>> (asl,_,_)=svp56_fast('speech.wav')
    """
    # if x is a filename, extact data from the wav file
    if isinstance(x, str):
        fs, x = scipy.io.wavfile(x)
    elif not isinstance(x, (list, np.ndarray)):
        raise ValueError("Invalid input for x")

    x = np.array(x)
    # only keep first channel in 2-channel audio
    if len(np.shape(x)) != 1:
        x = x[:, 0]

    # convert the signal to 16 bit integers in range [-32768, +32767]
    r = 2 ** 15
    # scale [-1, +1]
    if type(x[0]) == np.float32:
        x = np.asarray(x * r, dtype=np.int16)
    # scale [-2147483648, +2147483647]
    elif type(x[0]) == np.int32:
        x = np.asarray(x / r / 2, dtype=np.int16)
    # data is already in specified format
    elif type(x[0]) == np.int16:
        pass
    # scale [0, 255]
    elif type(x[0]) == np.uint8:
        offset = -((2 ** 8) - 1) / 2
        scale = (2 ** 8) + 1
        x = np.asarray((x + offset) * scale, dtype=np.int16)

    x = x - np.mean(x)  # mean removal is not in P56, but it should be

    r = 2 ** 15
    sq = np.sum(x ** 2)
    if sq == 0:
        raise ValueError("No signal")
    x = np.abs(x)  # rectify
    n = len(x)
    g = np.e ** (-1 / (fs * 0.03))

    q = sig.lfilter([(1 - g) ** 2], [1, -2 * g, g * g], x)  # smooth

    c = 2 ** np.arange(16)
    a = np.zeros(16)
    hs = round(0.2 * fs)  # 200 ms hang time

    # take log 2, with handling that gives log2(0)=-100 instead of error
    good = np.nonzero(0 < q)[0]
    lq = -100 * np.ones(len(q))
    lq[good] = np.floor(
        np.log2(q[good])
    )  # these 3 lines protect against log2(0), defining it to be log2(0)=-100
    # print(lq.tolist())
    for i in range(16):
        active = i <= lq  # compare log2 of env. with integer values 0 to 15
        trans = np.nonzero(np.abs(np.diff(active)))[0]
        for j in range(len(trans)):
            active[
                trans[j] : (min(trans[j] + hs - 1, n) + 1)
            ] = 1  # extend activity 200 ms
            # after each transition
        a[i] = np.sum(active)  # total number of log2(env) samples above a
        # threshold of i-1
    # ------------------Inspection point A-----------------------------

    L = (10 * np.log10(sq / n)) - (
        20 * np.log10(r)
    )  # average power relative to clip point (in dB)
    c = (20 * np.log10(c)) - (
        20 * np.log10(r)
    )  # binary thresholds relative to clip point (in dB)
    A = np.zeros(len(c))
    for i in range(16):
        if a[i] > 0:
            A[i] = 10 * np.log10(sq / a[i]) - 20 * np.log10(r)
        else:
            A[i] = 200
    d = A - c

    Afinal = 0
    for i in range(15):
        if d[i] >= 15.9 and d[i + 1] <= 15.9:
            if d[i] - d[i + 1] == 0:
                Afinal = A[i]
            else:
                s = (d[i] - 15.9) / (d[i] - d[i + 1])
                Afinal = (1 - s) * A[i] + (s * A[i + 1])

    saf = 10 ** ((L - Afinal) / 10)
    asl = Afinal

    # To get to a target active speech level, you can scale speech based
    # on the output of this code but you might have to iterate a bit to hit
    # an exact value.  This is because the
    # Speech Voltmeter is not exactly linear.  For example, scaling speech by
    # 2.0 does not produce exactly 20*log10(2)=6.02 dB increase in Speech
    # Voltmeter output. But it does typically produce a value in the range
    # 6.02 +/- 0.25 dB.

    # Here ends the code that emulates the ITU-T P.56 Speech Voltmeter.  Below is code to
    # create an activity mask.  It is the intuitive, reasonable extension of the above.
    # Specifically, if sum(a(i)) for some i goes to n, then we have 100# speech activity.
    # This says that the a(i)'s are the speech activities masks.  The a(i)'s are
    # calculated from floor(log2(envelope)).  The floor operation means that non-integral
    # thresholds on this "integer-log-envelope" are not meaningful.  Thus it remains
    # to simply pick the best i value and a(i) is the mask. We recaculate a(i) instead of
    # saving all 16, which could use a lot of memory for long speech files.

    dud = np.min(np.abs(d - 15.9))
    loc = np.argmin(np.abs(d - 15.9))
    active = loc <= lq
    trans = np.nonzero(np.abs(np.diff(active)))[0]
    for j in range(len(trans)):
        active[trans[j] : min(trans[j] + hs - 1, n) + 1] = 1
    return asl, saf, active


def load_cp(fname):
    """
    Read in cutpoints from file.

    Read cutpoints from the file with the given filename. Convert cutpoint
    indices from 1 based indices to zero based indices. Returns a tuple containing
    a dict for each segment of the file.

    Parameters
    ----------
    fname : str
        Name of cutpoint file to read.

    Returns
    -------
    tuple of dicts
        Tuple of dicts each with keys ['Clip','Start','End'].

    See Also
    --------
    mcvqoe.write_cp : Function to write cutpoints.

    Examples
    --------
    Example of loading cutpoints

    >>> load_cp('cp.csv')
    """
    # field names for cutpoints
    cp_fields = ["Clip", "Start", "End"]
    # open cutpoints file
    with open(fname, "rt") as csv_f:
        # create dict reader
        reader = csv.DictReader(csv_f)
        # check for correct fieldnames
        if reader.fieldnames != cp_fields:
            raise RuntimeError(f"Cutpoint columns do not match {cp_fields}")
        # create empty list
        cp = []
        # append each line
        for row in reader:
            # convert values to float
            for k in row:
                if k == "Clip":
                    # float is needed to represent NaN
                    row[k] = float(row[k])
                    # convert non nan fields to int
                    if not np.isnan(row[k]):
                        row[k] = int(row[k])
                else:
                    # make indexes zero based
                    row[k] = int(row[k]) - 1

            # append row to
            cp.append(row)
        return tuple(cp)


def write_cp(fname, cutpoints):
    """
    Write cutpoints to file.

    Write cutpoints from `cutpoints` to the file named `fname`.

    Parameters
    ----------
    fname : str
        Name of cutpoint file to read.
    cutpoints : tuple of dicts
        Tuple of dicts each with keys ['Clip','Start','End'].

    See Also
    --------
    mcvqoe.load_cp : Function to write cutpoints.

    Examples
    --------
    Example of saving cutpoints

    >>> cutpoints=({'Clip': 24, 'Start': 0, 'End': 41970})
    >>> write_cp('cp.csv',cutpoints)
    """
    # field names for cutpoints
    cp_fields = ["Clip", "Start", "End"]
    # open cutpoints file
    with open(fname, "wt", newline="\n", encoding="utf-8") as csv_f:
        # create dict writer
        writer = csv.DictWriter(csv_f, fieldnames=cp_fields)
        # write header row
        writer.writeheader()
        # write each row
        for wcp in cutpoints:
            # convert back to 1 based index
            wcp["Start"] += 1
            wcp["End"] += 1
            # write each row
            writer.writerow(wcp)


def a_weighted_power(x, fs=48000):
    """
    Calculate A-weighted power level for audio vector.

    Calculates an A-weighted power level in dB for the input audio vector x.
    This is a good approximation to relative loudness because the A-weighting
    function emulates a fundamental attribute of human hearing.

    The filter coefficients come from dsprelated.com and the resulting
    frequency response agrees with those shown in the literature.
    Getting filter coefficients for sample rates other than 48000 proved
    difficult so this code just converts input signals to 48000 instead.
    -S. Voran, Sept. 25, 2012

    Parameters
    ----------
    x : NumPy Array
        The audio in NumPy array format.
    fs : numeric, default=48000
        Sample rate for audio. Must be one of [8000, 16000, 24000, 32000,48000].

    Returns
    -------
    float
        A-weighted power for audio vector.

    See Also
    --------
    mcvqoe.svp56_fast : Speech Voltmeter function.

    Examples
    --------
    Calculate the A-weighted power for a 48 kHz audio vector

    >>> a_weighted_power(audio)
    """

    # Coefficients for A-weighting filter with fs=48000
    b = np.array(
        [
            0.01147155239724,
            -0.0248548824653166,
            0.0323052801494283,
            -0.0555571372813964,
            0.233784933167266,
            0.392882400411297,
            -0.633249911806281,
            -0.479494344932334,
            0.322061493940389,
            0.197659563091056,
            0.00299879461389451,
        ]
    )

    a = np.array(
        [
            1,
            -0.925699454182466,
            -0.992471193943543,
            0.837650096562845,
            0.22307303912603,
            -0.158404327757755,
            0.0184103295763937,
        ]
    )

    # Set gain at 1kHz to 0dB
    b = b * 1.1389

    # Resample audio
    if fs == 8000:
        x = sig.resample_poly(x, 6, 1)
    elif fs == 16000:
        x = sig.resample_poly(x, 3, 1)
    elif fs == 24000:
        x = sig.resample_poly(x, 2, 1)
    elif fs == 32000:
        x = sig.resample_poly(x, 3, 2)
    elif fs != 48000:
        raise ValueError(f"Unsupported sample rate of {fs}.\n")

    x = sig.lfilter(b, a, x)
    return 10 * (np.log10(np.mean(np.square(x))))

def get_measurement_from_file(filepath, module=True):
    """
    Determine measurement type from file header.

    Parameters
    ----------
    filepath : str
        Path to file.
    module : bool, optional
        If True returns the module name instead of evaluation shortened 
        measurement string (e.g. 'mcvqoe.mouth2ear' vs 'm2e'). The default is 
        True.

    Returns
    -------
    measurement : str
        Measurement type. Returns None if it could not be determined.

    """
    # Intialize measurement
    measurement = None
    try:
        # Read in csv
        df = pd.read_csv(filepath)
        # Get column names
        colnames = df.columns
        
        if 'W0_Int' in colnames:
            # All psud tests have to have at least one word so W0_Int should always be there
            measurement = 'psud'
        elif 'Intelligibility' in colnames:
            # Intelligibility only in intelligibility tests
            measurement = 'intell'
        elif 'ptt_st_dly' in colnames:
            # ptt_st_dly only in access time
            measurement = 'access'
        elif 'm2e_latency' in colnames:
            # m2e_latency in many tests, but if it's here and the others didn't work it's probably m2e latency
            measurement = 'm2e'
    except pd.errors.ParserError:
        # If parsing failed, check access data, skip 3 rows
        df = pd.read_csv(filepath, skiprows=3)
        if 'P1_Int' in df.columns:
            measurement = 'access'
        else:
            # Check if TVO data
            df = pd.read_csv(filepath, skiprows=2)
            if 'FSF' in df.columns:
                measurement = 'tvo'
        
            
    
    if module and measurement is not None:
        measurement_dir_modules = {    
            'access' : 'mcvqoe.accesstime',
            'intell' : 'mcvqoe.intelligibility',
            'm2e' : 'mcvqoe.mouth2ear',
            'psud' : 'mcvqoe.psud',
            'tvo': 'mcvqoe.tvo',
        }
        measurement = measurement_dir_modules[measurement]
    return measurement

        



from .Probabilityiser import PBI
from .QoEsim import ImpairmentParam
from numpy import inf

description = 'An impairment to represent a \'flakey\' channel. It is recommended to not use this as a pre impairment.'

impairment_type = 'generic'

parameters = {
    'P_a1'     : ImpairmentParam(1.0, float, 'range', min_val=0.0, max_val=1.0,
                                 interval = 0.001,
                                 description = 'Original probability transitioning into a transmitting state.'
                                ),
    'P_a2'     : ImpairmentParam(1.0, float, 'range', min_val=0.0, max_val=1.0, interval = 0.001,
                                 description = 'Probability transitioning into a transmitting state.',
                                ),
    'P_r'      : ImpairmentParam(1.0, float, 'range', min_val=0.0, max_val=1.0, interval = 0.001,
                                 description = 'Probability of remaining in a transmitting state.',
                                ),
    'interval' : ImpairmentParam(1.0, float, 'positive', interval = 0.001,
                                description = 'How often state transitions are evaluated.'
                                ),
}

def create_impairment(P_a1=1, P_a2=None, P_r=1, interval=1):

    impairment_class = PBI(P_a1=P_a1, P_a2=P_a2, P_r=P_r, interval=interval)
    
    return impairment_class.process_audio



import warnings

import numpy as np
import scipy.signal as sig
from numpy.fft import fft, ifft


def ITS_delay_est(x_speech, y_speech, mode, fs=8000, dlyBounds=[np.NINF, np.inf], min_corr=0):
    """
    Estimate the delay history for speech wave forms.

    This function estimates the delay history for the speech samples in
    y_speech relative to the speech samples in x_speech. Typically y_speech
    contains samples from the output of some system under test, and x_speech
    contains the corresponding input samples. Delay_est then holds estimates
    of the delay of that system under test.

    Written by Stephen Voran at the Institute for Telecommunication Sciences,
    325 Broadway, Boulder, Colorado, USA, svoran@its.bldrdoc.gov
    May 10,2004

    Parameters
    ----------
    x_speech : array
        Vector of speech samples. At least 148 ms of speech is required.
    y_speech : array
        Vector of speech samples. At least 148 ms of speech is required.
    mode : {'f','v','u'}
        Delay mode to use. Fixed delay ('f') assumes that the delay is fixed
        for all of y_speech. Variable delay ('v') assumes that the delay can
        change. Unknown delay ('u') checks if the coarse delay was well
        correlated, fixed delay is used, otherwise variable delay is used.
    fs : numeric, default=8000
        Sample rate for the audio vectors.
    dlyBounds : array, default = [-Inf,Inf]
        The interval of acceptable delays in seconds.
    min_corr : float, default = 0
        Minimum correlation threshold. If the coarse delay correlation is lower
        than min_corr, the delay estimate is determined to be unsuccessful.

    Returns
    -------
    pos : int or tuple of ints
        End of each delay estimate segment in samples. If mode is 'f' this will
        be an int (only one delay) otherwise pos will be a tuple of ints.
    dly : int or tuple of ints
        Delay estimates corresponding to the segments in pos. If mode is 'f'
        this will be an int, otherwise dly will be a tuple of ints the same
        length as pos.

    See Also
    --------
    mcvqoe.delay.sliding_delay_estimates : Uses ITS_delay_est.
    mcvqoe.base.audio_read : Function for reading in audio data.

    Examples
    --------
    load in speech file for examples

    >>> speechf=io.BytesIO(pkgutil.get_data('mcvqoe','audio_clips/test.wav'))
    >>> (fs_speech,speech)=mcvqoe.base.audio_read(speechf)

    Simple fixed delay example.

    >>> fixed_delay_speech=np.concatenate((np.zeros(shape=[17]),speech))
    >>> mcvqoe.ITS_delay_est(speech,fixed_delay_speech,'f')
    (161635, 17)

    This means that the single delay estimate of 17 samples applies to samples
    with indices [0,161635] in y_speech.

    >>> var_delay_speech=np.concatenate((speech[2:12000],np.zeros(shape=[417]),
    ...                                  speech[12015:]))
    >>> mcvqoe.ITS_delay_est(speech,var_delay_speech,'u')
    ((12184, 162018), (-2, 400))

    This means that for samples 0 through 12184, the delay is estimated to be -2
    samples. For samples 12184 through 162018, the delay is estimated to be
    400 samples.

    If 'u' mode is used, a tuple of tuples is always returned, even if one delay
    is found.

    >>> mcvqoe.ITS_delay_est(speech,fixed_delay_speech,'u')
    ((161635,), (17,))

    Example with too small an input signal.

    >>> ITS_delay_est(speech,speech[:1000],'f')
    array([0, 0])
    """
    # ----------------------------Parse Arguments--------------------------------
    x_speech = np.array(x_speech, dtype=np.float64)
    if len(x_speech) == 0:
        raise ValueError("x_speech can not have zero length")
    if x_speech.ndim != 1:
        raise ValueError(f"Expected 1 dimension for x_speech but {x_speech.ndim} found")

    y_speech = np.array(y_speech, dtype=np.float64)
    if len(y_speech) == 0:
        raise ValueError("y_speech can not have zero length")
    if y_speech.ndim != 1:
        raise ValueError(f"Expected 1 dimension for y_speech but {y_speech.ndim} found")

    if mode not in ("f", "v", "u"):
        raise ValueError(f'Unknown mode "{mode}", expected mode to be one of "f", "v", or "u"')

    dlyBounds = np.array(dlyBounds, dtype=np.float64)
    if len(dlyBounds) != 2:
        raise ValueError(f"Expected dlyBounds len to be 2 but got {len(dlyBounds)}")
    elif dlyBounds[1] <= dlyBounds[0]:
        raise ValueError(f"dlyBounds must be increasing. got {dlyBounds}")

    # ----------------------Resample Arguments to 8kHz--------------------------
    if fs != 8000:
        xlen = len(x_speech)
        ylen = len(y_speech)
        x_speech = sig.resample(x_speech, int(xlen * 8000 / fs))
        y_speech = sig.resample(y_speech, int(ylen * 8000 / fs))

    # ---------------------------Check return type------------------------------#

    if mode == "f":
        ret_type = "single-value"
    else:
        ret_type = "multiple-values"

    # --------------------------Level Normalization-----------------------------#
    # Measure active speech level
    try:
        asl_x = active_speech_level(x_speech)
        asl_y = active_speech_level(y_speech)
    except ValueError:
        warnings.warn("Input vector has no signal")
        Delay_est = np.array([0, 0], ndmin=2)

        if ret_type == "single-value":
            # return results as a tuple of ints
            return (int(Delay_est[:, 0]), int(Delay_est[:, 1]))
        else:
            # return results as a tuple of tuples
            return (tuple(Delay_est[:, 0]), tuple(Delay_est[:, 1]))
    # Force active speech level to -26 dB r.e. overload
    x_speech = x_speech * 10 ** ((asl_x + 26) / -20)
    y_speech = y_speech * 10 ** ((asl_y + 26) / -20)
    # ---------------------Coarse Average Delay Estimation----------------------#
    tau_0, rho_0, fir_coeff_63 = coarse_avg_dly_est(x_speech, y_speech, dlyBounds)
    # Compensate for tau_0, comp_x_speech and comp_y_speech will have same length
    comp_x_speech, comp_y_speech = fxd_delay_comp(x_speech, y_speech, tau_0)
    # If compensation for tau_0 results in a speech vector that is shorter
    # than 1185 samples, then the input signal vectors x_speech and y_speech
    # are not sufficent for delay estimation. (They may contain unrelated
    # signals, no signal, or may simply be too short)
    if len(comp_x_speech) < 1185:
        # Algorithm must terminate
        mode = "t"
    # ------------------------Check Correlation threshold-----------------------#
    if rho_0 < min_corr:
        # rho_0 is too low, terminate
        mode = "t"
    # ------------Do further mode determination as necessary/possible-----------#
    if mode == "u" and rho_0 < 0.96:
        mode = "v"
    # -----Fine delay estimation for the fixed and unknown delay cases----------#
    if mode == "f" or mode == "u":
        # Find fine delay
        fxd_fine_delay = fxd_fine_dly_est(comp_x_speech, comp_y_speech)
        # Find total delay
        D_fxd = tau_0 + fxd_fine_delay

        # check if we are in fixed mode (not 'u')
        if mode == "f":
            # nothing more to do, bail early
            return (int((len(y_speech) - 1) * (fs / 8000)), int(D_fxd * (fs / 8000)))
    # -------Additional stages for the variable and unknown delay cases---------#
    if mode == "v" or mode == "u":
        # ---------------------Speech Activity Detection------------------------#
        # Identify active speech samples (active_wf is same size as y_speech,
        # 1 indicates activity, 0 otherwise)
        active_wf = find_activity_wf(y_speech, fir_coeff_63)
        # Compensate the activity waveform for tau_0
        _, comp_active_wf = fxd_delay_comp(x_speech, active_wf, tau_0)
        # --------------------------Delay Tracking------------------------------#
        DCAVS = delay_tracking(comp_x_speech, comp_y_speech, comp_active_wf, 150, 40, 200)
        # --------------------------Median Filtering----------------------------#
        SDV = median_filter(DCAVS, 500, 40, 0.1, 0.8)
        # ---------------------Combine Results with tau_0-----------------------#
        SDV[:, 1] = SDV[:, 1] + tau_0  # Add in tau_0 to delay estimates
        if 0 < tau_0:  # If tau_0 is a positive quantity
            # then adjust locations of delay segments as well
            SDV[:, 0] = SDV[:, 0] + tau_0
        # Adjust final location to exactly match end of y_speech
        SDV[-1, 0] = len(y_speech) - 1
        # ----------------------------Delay Refinement--------------------------#
        SDV = delay_refine(SDV, x_speech, y_speech, active_wf, 72, 0.7)
        # ---------------Remove Redundant Entries in SDV matrix-----------------#
        # If delay and validity do not change from segment n to n+1, then
        # segment n is redundant
        keepers = np.append(
            np.nonzero(np.logical_or(np.diff(SDV[:, 1]) != 0, np.diff(SDV[:, 2]) != 0))[0],
            np.size(SDV, 0) - 1,
        )
        SDV = SDV[keepers, :]
        # Adjust final location to exactly end y_speech
        SDV[-1, 0] = len(y_speech) - 1
        # -----------------Round Delay Estimates to Nearest Integer-------------
        SDV[:, 1] = np.round(SDV[:, 1])
        # -----------------------Short Segment Correction------------------------
        if np.size(SDV, 0) > 1:
            SDV = short_seg_cor(SDV, x_speech, y_speech, 160, 280, 80)

    # --------------------------Apply LSE if Necessary--------------------------
    if mode == "u":
        [lse_f, lse_v] = LSE(x_speech, y_speech, D_fxd, SDV, 16)

        if lse_f <= lse_v:
            mode = "f"
        else:
            mode = "v"
    # ---Select Output, Extrapolate Variable Delay Estimate if Necessary--------
    if mode == "v":
        Delay_est = extend_val_res(SDV)  # Extrapolate variable delay estimate
    elif mode == "f":
        Delay_est = np.array([[len(y_speech) - 1, D_fxd]])  # Reformat fixed delay estimate
    else:  # Mode is 't' for terminate
        # The output [0 0] indicates no delay estimation was possible
        Delay_est = [0, 0]
    # convert to input sample rate and truncate to integer
    Delay_est = (np.array(Delay_est, ndmin=2) * (fs / 8000)).astype(int)

    if ret_type == "single-value":
        # return results as a tuple of ints
        return (int(Delay_est[:, 0]), int(Delay_est[:, 1]))
    else:
        # return results as a tuple of tuples
        return (tuple(Delay_est[:, 0]), tuple(Delay_est[:, 1]))
    # ==========================================================================


def active_speech_level(x, fs=8000):
    # Usage: asl=active_speech_level(x)
    # This function measures the active speech levels in the speech vector x.
    # x is a vector of speech samples
    # fs to is the sample rate of x
    # asl is the active speech level in dB relative to overload
    # code will extend each active region by tau samples (forward in time)
    tau = round(0.200 * fs)
    # active speech is defined to be dBth dB below max
    dBth = 20
    n = len(x)
    x = x - np.mean(x)  # mean removal
    x = np.abs(x)
    # calculate filter coefficient from time constant
    g = np.exp(-1 / (fs * 0.03))
    # perform 2nd order IIR filtering
    x = sig.lfilter([(1 - g) ** 2], [1, -2 * g, g * g], x)
    at = max(x) * (10 ** (-dBth / 20))  # calculate activity threshold

    if at == 0:
        raise ValueError("Input vector has no signal")
    active = x > at  # find active samples
    # Extend each active interval tau samples forward in time
    trans = np.nonzero(np.abs(np.diff(active)))[0]

    for i in range(0, len(trans)):
        active[trans[i] : min(trans[i] + tau, n - 1) + 1] = 1

    # Test for both activity and non-zeroness to prevent log(0)
    x = x[np.logical_and(0 < x, active)]
    asl = 20 * np.mean(np.log10(x)) - 81
    return asl


# ==========================================================================
def find_fir_coeffs(order, cutoff):
    # Usage: b=find_fir_coeffs(order,cutoff)
    # This function generates filter coefficients for an FIR LPF with
    # the specified order and cutoff frequency. Cutoff frequency
    # is specified relative to Nyquist frequency. b is a row vector
    # of length order+1 that holds the filter coefficients
    # Create vector that holds Hamming window of length order+1
    n = order + 1
    h = 0.54 - 0.46 * np.cos(2 * np.pi / (n - 1) * np.arange(0, n))
    # Create vector of time-domain indices
    t = np.arange(-order * cutoff / 2, (order * cutoff / 2) + cutoff, cutoff)
    # Calculate sin(pi*x)/(pi*x) (sinc function) for time domain indices
    # sin(pi*0)/(pi*0) is defined to be 1
    good = t != 0
    sinxox = np.ones(order + 1)

    sinxox[good] = np.sin(np.pi * t[good]) / (np.pi * t[good])
    # np.sin(np.pi * t[good])/(np.pi*t[good])
    # Filter coefficients are product of window and sinc function
    b = h * sinxox
    # Normalize coefficients for unity gain in passband
    b = b / sum(b)
    return b


# ==========================================================================
def coarse_avg_dly_est(x, y, b):
    # Usage: tau_0,rho_0,fir_coeff=coarse_avg_dly_est(x,y, b)
    # This function generates a coarse delay estimate from speech envelopes.
    # This is done in the fs=125 samples/sec domain. (i.e., sub-sampling by 64).
    # x and y are vectors of speech samples
    # b is tuple of a lower and lower bound of acceptable delays in seconds
    # tau_0 is a delay estimate in samples
    # rho_0 is the corresponding correlation value
    # fir_coeff is a set of 401 FIR filter coefficients for a 63 Hz LPF
    fir_coeff = find_fir_coeffs(400, 1 / 133.33)  # calculate filter coeffs
    ex = sig.lfilter(fir_coeff, 1, np.abs(x))  # 63 Hz LPF, order 400 FIR
    ey = sig.lfilter(fir_coeff, 1, np.abs(y))
    # There is no need to remove filter delay since it is the same in each
    # signal
    ex = ex[0::64]  # Subsample by 64
    ey = ey[0::64]
    # zero pad so ex and ey have same length
    lx = len(ex)
    ly = len(ey)
    L = max(lx, ly)
    ex = np.append(ex, np.zeros(L - lx))
    ey = np.append(ey, np.zeros(L - ly))
    corrlen = len(ex)

    m = np.mean(ex)
    # Remove mean of ex from both ex and ey
    ex = ex - m
    ey = ey - m
    # FFT based cross correlation
    N = 2 * len(ex)
    term1 = fft(np.append(ex, np.zeros(corrlen)))
    term2 = fft(np.append(ey[::-1], np.zeros(corrlen)))

    xc = np.real(ifft(term1 * term2))
    # calculate shifts in seconds
    shift = 64 * (corrlen - np.arange(1, len(xc) + 1))
    # calculate which shifts are valid
    valid = np.logical_and(shift > (b[0] * 8e3), shift < (b[1] * 8e3))
    # calculate which shifts are valid
    valid_shifts = shift[valid]

    check = xc[valid]
    rho = max(check)
    index = np.argmax(check)
    # Convert peak location to a shift
    tau_0 = valid_shifts[index]
    # Normalize to get cross correlation value
    rho_0 = rho / ((corrlen - 1) * np.std(ex, ddof=1) * np.std(ey, ddof=1))
    return tau_0, rho_0, fir_coeff


# ==========================================================================
def fxd_delay_comp(source, distorted, delay):
    # Usage: source,distorted=delay_comp(source,distorted,delay)
    # This function compensates the source-distorted signal pair by the given
    # delay value in samples. The returned source and distorted signals will
    # have the same length.
    sstart = max(0, -delay)  # source starting point
    dstart = max(0, delay)  # distorted starting point
    # Number of samples available
    samples = min(len(source) - sstart, len(distorted) - dstart)
    # Extract proper portions
    source = source[sstart : sstart + samples]
    distorted = distorted[dstart : dstart + samples]
    return source, distorted


# ==========================================================================
def fxd_fine_dly_est(x, y):
    # Usage: D=fxd_fine_dly_est(x,y)
    # This function performs an FFT-based cross correlation on the rectified
    # speech signals and then processes the results to find a delay estimate
    # x and y are  vectors of speech samples
    # D is the estimated delay
    ran = 128  # half width of range of samples to analyze in this stage
    # number of samples to feed into filter while waiting for it to stabilize
    headlen = 500
    # number of samples (a tail) long enough to cover the longest filter delay
    taillen = 200
    # find cross correlation of rectified speech
    xc, denom = fft_xc(np.abs(x), np.abs(y), -(ran + headlen), ran + taillen)
    # extract relevant portion
    txc = xc[headlen : headlen + 1 + 2 * ran]
    # find max
    maxrho = max(txc)
    index = np.argmax(txc)

    maxrho = maxrho / denom
    if 0.73 < maxrho:  # For high correlations, no smoothing is required
        D = index - ran  # Calculate delay estimate
    else:
        if 0.67 < maxrho:  # For medium correlations, some smoothing helps
            m = 64
        else:  # For lower correlations, more smoothing helps
            m = 128
        flen = 3 * m
        fir_coeff = find_fir_coeffs(flen, 1 / m)  # Filter lengths are even
        sxc = sig.lfilter(fir_coeff, 1, xc)
        # smoothed version of cross-correlation function with filter delay
        # removed
        sxc = sxc[int(headlen + (flen / 2)) : int(headlen + (flen / 2) + 1 + 2 * ran)]
        index = np.argmax(sxc)
        D = index - ran  # Calculate delay estimate
    return D


# ==========================================================================
def fft_xc(x, y, min_d, max_d):
    # Usage: un_corr,denom=fft_xc(x,y,min_d,max_d)
    # This function does an fft-based cross correlation on the waveforms in the
    # vectors x and y. These two vectors need not have the same length,
    # but the resulting delay estimated is defined relative to zero time offset
    # of x(1) and y(1).
    #
    # If possible, an unnormalized correlation value is returned for all delay
    # values from min_d to max_d, inclusive. (If this is not possible, it is
    # an error condition)
    #
    # Thus un_corr is a length max_d-min_d+1 vector. When it is divided
    # by denom (typically after finding a max), correlation values result.
    # un_corr(1) corresponds to min_d, un_corr(end) corresponds to max_d
    # The legal search window is approximately +/- min(length(x),length(y).
    # Pad with zeros so x and y have same length
    lx = len(x)
    ly = len(y)
    L = max(lx, ly)
    x = np.append(x, np.zeros(L - lx))
    y = np.append(y, np.zeros(L - ly))
    corrlen = len(x)
    # Remove the mean of x from each signal
    m = np.mean(x)
    x = x - m
    y = y - m
    # Perform FFT-based cross correlation
    term1 = fft(np.append(x, np.zeros(corrlen)))
    term2 = fft(np.append(y[::-1], np.zeros(corrlen)))
    xc = np.real(ifft(term1 * term2))
    xc = xc[::-1]  # reverse the vector
    # Test to see if requested values of delay are available
    if corrlen + min_d + 1 < 1 or 2 * corrlen < corrlen + max_d + 1:
        error("Not enough input samples to calculate requested delay values.")
    # Extract requested values of delay
    un_corr = xc[corrlen + min_d : corrlen + max_d + 1]
    # Calculate the denominator
    denom = (corrlen - 1) * np.std(x, ddof=1) * np.std(y, ddof=1)
    return un_corr, denom


# ==========================================================================
def find_activity_wf(x, fir_coeff):
    # Usage: active_x=find_activity_wf(x,fir_coeff)
    # This function generates an output vector (active_x) that shows the
    # speech activity waveform of the input vector (x).
    # Speech activity is defined via a smoothed speech envelope with nominal
    # bandwidth of 63 Hz. The required FIR filter coefficients are given in
    # fir_coef. Samples of the envelope that are within 35 dB of
    # the envelope peak are associated with active speech. Each interval of
    # activity is extended by tau samples forward and backward in time.
    th = 35  # Threshold for activity detection in dB below peak
    tau = 800  # Number of samples to extend each active segment in each direction
    nx = len(x)
    # FIR filter rectified speech
    x = sig.lfilter(fir_coeff, 1, np.append(np.abs(x), np.zeros(200)))
    x = x[200:]  # Remove filter delay
    # Find samples that are above threshold,1 means active, 0 means not active
    active_x = x >= 10 ** (th / 20.0)
    # Extend all active regions by tau samples in each direction
    trans = np.nonzero(np.abs(np.diff(active_x)))[0]  # List of transition points
    for j in range(0, len(trans)):  # Loop over all transitions
        # Extend activity in each direction
        active_x[max(trans[j] - tau, 0) : min(trans[j] + tau, nx - 1) + 1] = 1
    return active_x


# ==========================================================================
def delay_tracking(x, y, active_wf, winlen, winstep, ran):
    # Usage: DCAVS=delay_tracking(x,y,active_wf,winlen,winstep,range)
    # This function does delay tracking (delay in speech signal y relative to x)
    # winlen is the length of window used for the delay estimation alg (in ms)
    # winstep is the step size between windows (in ms)
    # range is the half-width of the search range (in ms)
    # DCAVS is a nwins by 5 matrix, where nwins is the number of delay
    # estimation windows that can be fit into y
    # Column 0 holds Delay values (in the 16:1 subsampled domain)
    # Column 1 holds the corresponding Correlation values
    # Column 2 holds the corresponding Activity levels values
    # Column 3 holds the corresponding Validities (True for valid delay estimate,
    # False otherwise)
    # Column 4 holds the Sample number in y (in the 16:1 subsampled domain)
    # that is at the center of each window
    # Create coefficients for an FIR LPF with 129 taps and a delay of
    # 64 samples. Response is -51 dB at 250 Hz.
    fir_coeffs = find_fir_coeffs(128, 1 / 32)
    # Rectify and filter
    x = sig.lfilter(fir_coeffs, 1, np.append(np.abs(x), np.zeros(64)))
    # Remove filter delay and subsample by 16
    x = x[64::16]
    nx = len(x)
    # Rectify and filter
    y = sig.lfilter(fir_coeffs, 1, np.append(np.abs(y), np.zeros(64)))
    # Remove filter delay and subsample by 16
    y = y[64::16]
    # Subsample activity waveform
    active_wf = active_wf[0::16]
    # Force subsampled activity waveform to have same length as subsampled y
    ny = len(y)
    len_diff = len(active_wf) - ny
    if 0 < len_diff:
        active_wf = active_wf[0:ny]  # Trim final samples
    elif len_diff < 0:
        active_wf = np.append(active_wf, np.zeros(len_diff))  # Zero pad
    # Convert parameters from ms to samples (in the subsampled domain)
    winlen = round(winlen / 2)
    winstep = round(winstep / 2)
    ran = round(ran / 2)
    # Find total number of window positions that can be placed on y
    nwins = np.floor((ny - winlen) / winstep) + 1
    DCAVS = np.zeros((int(nwins), 5))
    # Find locations of centers of windows (sample number in subsampled domain)
    first = (winlen + 1) / 2 - 1
    last = first + winstep * (nwins - 1)
    DCAVS[:, 4] = np.arange(first, last + winstep, winstep)
    for i in range(0, int(nwins)):  # Loop over all windows
        start = i * winstep  # First sample in window
        stop = start + winlen  # Last sample in window
        # Find activity level in window
        DCAVS[i, 2] = sum(active_wf[start:stop]) / winlen
        # Assume delay estimation is doable unless one of tests that
        # follows fails
        doable = True
        if start - ran < 0 or min(nx, ny) < stop + ran:
            # Not enough samples to do delay estimation
            doable = False
        elif np.std(y[start : stop + 1], ddof=1) == 0 or np.std(x[start - ran : stop + ran + 1], ddof=1) == 0:
            # No variation in samples
            doable = False
        if doable:
            # Perform direct-form cross correlation
            xc, denom = non_fft_xc(x[start - ran : stop + ran], y[start - ran : stop + ran], -ran, ran)
            # Identify peak
            maxrho, index = np.max(xc), np.argmax(xc)
            # Calculate corresponding delay in samples
            # USED TO BE  A MINUS ONE HERE
            DCAVS[i, 0] = index - ran
            # Calculate corresponding correlation value
            DCAVS[i, 1] = maxrho / denom
            # Mark that window as having a valid delay estimate
            DCAVS[i, 3] = True
    return DCAVS


# ==========================================================================


def non_fft_xc(x, y, min_d, max_d):
    # Usage: xcs,denom,ystart,ystop=non_fft_xc(x,y,min_d,max_d)
    # This function enables delay estimation by calculating the cross
    # correlation between two vectors of speech samples x and y at the
    # specified shifts. x and y need not have the same length. Zero delay
    # is associated with the case where x(1) aligns with y(1). x and y may
    # be row or vectors. Cross correlations are performed at all
    # delays between min_d and max_d (given in samples) inclusive. If the
    # length of x or y prevents this, an error is generated.
    # xcs is a vector and holds unnormalized correlation values for all
    # requested delays
    # denom is the normalization factor so that xcs/denom will be true
    # correlation values.
    # Note that length(xcs)=max_d-min_d+1. xcs(1) is associated with min_d,
    # xcs(end) is associated with max_d.
    # Note also that this function always uses a fixed segement of samples of y.
    # The number of samples in this fixed segment is maximized given the
    # constraints imposed by the lengths of x and y as well as the values of
    # min_d and max_d.
    # ystart and y stop are the first and last samples of y that are used in
    # the fixed segment.
    # Find number of shifts
    nshifts = max_d - min_d + 1
    # Initialze correlation results variable
    xcs = np.zeros(nshifts)
    # Find two lengths
    nx = len(x)
    ny = len(y)
    # Find segment of y that can be used
    ystart = max(0, max_d)  # ystart is first sample of y to use
    if ny - min_d < nx:
        ystop = ny  # ystop is last sample of y to use
    else:
        ystop = nx + min_d
    # Generate error if there is no useable segment of y
    if ystop < ystart:
        raise ValueError("Not enough input samples to calculate all delay values.")
    # Extract segment of y
    temp_y = y[ystart:ystop]
    # Find number of samples in segment
    m = ystop - ystart
    # Loop over all shifts
    for i in range(0, nshifts):
        # Update current delay value
        cd = min_d + i
        xstart = ystart - cd
        # Extract proper segment of x
        temp_x = x[xstart : xstart + m]
        # Form partial denominator of correlation so it can be tested
        # to prevent divide by zero
        denom = np.dot(temp_x, temp_x)
        if denom > 0:
            # Correlation
            xcs[i] = np.dot(temp_x, temp_y) / np.sqrt(denom)
        else:
            # When segment of x has no signal, correlation is zero
            xcs[i] = 0
    # Find the fixed portion of the denominator
    denom = np.sqrt(np.dot(temp_y, temp_y))
    return xcs, denom


# =========================================================================='''
def median_filter(DCAVS, twinlen, winstep, activity_th, cor_th):
    # Usage: SDV=median_filter(DCAVS,twinlen,winstep,activity_th,cor_th)
    # This function does the median filtering on the results in the DCAVS
    # matrix. The DCAVS matrix is defined in the delay_tracking function.
    #
    # twinlen is the length of the median filtering window in ms
    # winstep tell how many ms each step in DCAVS matrix corresponds to worth
    # activity_th is the activity threshold required for a sample to be included
    # in the median filter
    # cor_th is the correlation threshold required for a sample to be for
    # included in the median filtering
    #
    # Results are returned in the matrix SDV. Each row of SDV describes a
    # segment of constant delay. SDV is in the fs=8000 domain.
    # Column 0, Sample number of last sample of constant delay segment
    # Column 1, estimated Delay of segment
    # Column 2, Validity of delay estimate (0 for invalid, 1 for valid)
    #
    # SDV(end,1) is given by the center of the final window (DCAVS(end,5)),
    # converted from the fs=500 domain to the fs=8000 domain. Note that in
    # general, this will not be exactly the same as the length of the
    # speech signal y (output from system under test).
    #
    # If no valid information can be extracted from DCAVS, the result is a delay
    # of zero everywhere.
    # Extract number of time samples available
    nwins = np.shape(DCAVS)[0]
    # Create a temporary matrix. It has 1 row for each row of DCAVS.
    # Column 0 will hold delay estimates
    # Column 1 will hold True if that estimate is valid, False otherwise
    # Column 2 will hold the distorted speech envelope sample number
    # (in the fs=500 domain) associated with the delay estimate
    T = np.column_stack((np.zeros((nwins, 2)), DCAVS[:, 4]))
    # Find half-width of median filtering window in samples
    htwinlen = round(twinlen / (2 * winstep))
    # Good is a vector with length nwins. It has a 1 where a
    # delay estimate has correlation that meets or exceeds threshold,
    # activity that meets or exceeds threshold, and is mathematically valid.
    # It has a zero elsewhere.
    good = np.logical_and.reduce((cor_th <= DCAVS[:, 1], activity_th <= DCAVS[:, 2], DCAVS[:, 3]))
    # Loop over all samples
    for i in range(0, nwins):
        # Check number of samples between sample and last sample
        nsmp = min(i, nwins - i - 1)
        # Find final half-width of median filtering window (cannot exceed the
        # number of remaining samples)
        fhtwinlen = min(htwinlen, nsmp)
        start = i - fhtwinlen  # First sample in current window
        stop = i + fhtwinlen + 1  # Last sample in current window
        # If there is a least one good sample in the window
        if sum(good[start : stop + 1]) > 1:
            # Form list of absolute indices of good samples
            goodlist = start + np.nonzero(good[start:stop])[0]
            # Perform median filtering on the good samples
            # Note on median function: When presented with an even number of
            # samples, this median function returns the average of the two
            # central samples. e.g. median([1 2 3 4])=2.5
            T[i, 0] = np.median(DCAVS[goodlist, 0])
            # Mark that a valid result has been calculated
            T[i, 1] = True
    # Find how many valid results have been calculated
    nvalres = sum(T[:, 1])
    # If no valid results have been calculated, report zero delay everywhere
    if nvalres == 0:
        SDV = [DCAVS[:, 4], 0, 0]
    else:
        # List only results that describe a change in delay or a change in
        # validity, plus the final result
        keepers = np.append(
            np.nonzero(np.logical_or(np.diff(T[:, 0]) != 0, np.diff(T[:, 1]) != 0)),
            nwins - 1,
        )
        # Keep only those results
        SDV = np.column_stack((T[keepers, 2], T[keepers, 0], T[keepers, 1]))
    # Convert final results from fs=500 samples/sec domain to the
    # fs=8000 samples/sec domain
    SDV[:, 1] = SDV[:, 1] * 16  # Convert delay estimates
    SDV[:, 0] = (SDV[:, 0]) * 16 + 8  # Convert sample values
    return SDV


# ==========================================================================


def delay_refine(SDVin, x_speech, y_speech, active_wf, ran, cor_th):
    # Usage: SDVout=delay_refine(SDVin,x_speech,y_speech,active_wf,range,cor_th)
    # This function refines the input delay matrix SDVin, to generate SDVout.
    # SDVin and SDVout are in the fs=8000 domain and have one row per segment
    # of constant delay.
    # Column 0, Sample number of last sample of constant delay segment
    # Column 1, estimated Delay of segment
    # Column 2, Validity of delay estimation segment (0 for invalid, 1 for valid)
    #
    # x_speech and y_speech are vectors of speech samples
    # with sample rate 8000 samples/sec. (x_speech is associated with system
    # under test input, y_speech is associated with system under test output)
    # active_wf has the same size as y_speech. 1 indicates speech activity,
    # 0 otherwise.
    # range gives the search range for this stage in samples
    # cor_th is the correlation threshold for refinement
    # Rectify speech
    x_speech = np.abs(x_speech)
    y_speech = np.abs(y_speech)
    # Copy all data, function will refine it where possible
    SDVout = SDVin
    # Find number of segments of constant delay
    nsegs = np.size(SDVin, axis=0)
    # Loop over all segments
    for i in range(0, nsegs):
        # Find first sample of current segment (first segment is special case)
        if i == 0:
            start = 0
        else:
            start = int(SDVin[i - 1, 0] + 1)
        # Extract last sample of current segment
        stop = int(SDVin[i, 0])
        # Extract delay of current segment
        delay = SDVin[i, 1]
        # Attempt refinement only if there is at least 10 ms of active speech
        # in the segment and segment has a valid delay estimate
        if 80 <= sum(active_wf[start : stop + 1]) and SDVin[i, 2] == 1:
            # If segment length is at least 200 ms, use fft-based correlation
            if 1600 <= (stop - start + 1):
                # Find delay compensated starting sample in x_speech
                sstart = start - delay
                # If it is before the start of x_speech, it will be necessary to
                # modify the starting place in both x_speech and y_speech
                if sstart < 0:
                    # Number of samples involved in the modifications
                    trim = -sstart
                    # First sample of x_speech for this segement (index 0)
                    sstart = sstart + trim
                    # First sample of y_speech for this segement
                    start = start + trim
                # Find delay compensated ending sample in x_speech
                sstop = min(stop - delay, len(x_speech) - 1)
                # If there is at least 10 ms in both the x_speech segment and
                # the y_speech segment after these adjustments
                if (80 <= sstop - sstart + 1) and (80 <= stop - start + 1):
                    # Perform FFT-based cross correlation on appropriate
                    # portions of x_speech and y_speech
                    sstart = int(sstart)
                    start = int(start)
                    stop = int(stop)
                    sstop = int(sstop)
                    un_corr, denom = fft_xc(
                        x_speech[sstart : sstop + 1],
                        y_speech[start : stop + 1],
                        -ran,
                        ran,
                    )
                    # Locate peak in unnormalized correlation
                    peak = np.max(un_corr)
                    loc = np.argmax(un_corr)
                    # If correlation value meets or exceeds threshold, or the
                    # segment is longer than 1 second
                    if (cor_th <= peak / denom) or (8000 < stop - start + 1):
                        # Apply the refinement
                        SDVout[i, 1] = delay + (loc - ran)
                # Segment is less than 200 ms long, use direct-form correlation
            else:
                # Find starting sample in x_speech, compensated for
                # delay and search range
                sstart = start - delay - ran
                # If it is before the start of x_speech, it will be necessary to
                # modify the starting place in both x_speech and y_speech
                if sstart < 0:
                    # Number of samples involved in the modifications
                    trim = -sstart
                    # First sample of x_speech for this segement
                    sstart = sstart + trim
                    # First sample of y_speech for this segement
                    start = start + trim
                # Find ending sample in x_speech, compensated for
                # delay and search range
                sstop = stop - delay + ran
                # If it is beyond the end of x_speech, it will be necessary to
                # modify the ending place in both x_speech and y_speech
                if len(x_speech) < sstop:
                    # Number of samples involved in the modifications
                    trim = sstop - len(x_speech)
                    # Last sample of x_speech for this segement
                    sstop = sstop - trim
                    # Last sample of y_speech for this segement
                    stop = stop - trim
                # If there is at least 10 ms in the y_speech segment after
                # these adjustments
                if 80 < stop - start + 1:
                    sstart = int(sstart)
                    start = int(start)
                    stop = int(stop)
                    sstop = int(sstop)
                    # Perform direct-form cross correlation on appropriate
                    # portions of x_speech and y_speech
                    un_corr, denom = non_fft_xc_all(x_speech[sstart : sstop + 1], y_speech[start : stop + 1])
                    # Locate peak in unnormalized correlation
                    peak = max(un_corr)
                    loc = np.argmax(un_corr)
                    # If correlation value meets or exceeds threshold
                    if cor_th <= peak / denom:
                        # Apply the refinement
                        SDVout[i, 1] = delay + (ran - loc)
    return SDVout


# ==========================================================================
def non_fft_xc_all(x, y):
    # This function applies a direct-form cross correlation to the
    # vectors of speech samples x and y. This correlation is done for all
    # possible shifts that use all of y.
    # The unnormalized correlation values are returned in un_cor, and the
    # denominator is returned in denom.
    # It is required that length(x)>=length(y).
    # Find lengths
    nx = len(x)
    ny = len(y)
    # Find number of possible shifts
    nshifts = nx - ny + 1
    # Initialize results variable
    un_corr = np.zeros(nshifts)
    # Loop over all possible shifts
    for i in range(0, nshifts):
        temp_x = x[i : i + ny]
        # Form partial denominator of correlation so it can be tested
        # to prevent divide by zero
        denom = np.dot(temp_x, temp_x)

        if denom > 0:
            # Correlation
            un_corr[i] = np.dot(temp_x, y) / np.sqrt(denom)
        else:
            # When segment of x has no signal, correlation is zero
            un_corr[i] = 0
    # Find fixed portion of denominator
    denom = np.sqrt(np.dot(y, y))
    return un_corr, denom


# ==========================================================================
def short_seg_cor(SDVin, x_speech, y_speech, len_t, len_b, len_s):
    # Usage: SDVout=short_seg_cor(SDVin,x_speech,y_speech,len_t,len_b,len_s)
    # This function tests all pulses (also called blips), steps and tails in
    # an estimated delay history and removes them when appropriate
    #
    # x_speech and y_speech are are vectors that hold system under test
    # input and output speech samples (without any delay compensation)
    # len_t is the length in ms of the longest tail that should be removed
    # len_b is the length in ms of the longest blip that should be removed
    # len_s is the length in ms of the longest step that should be removed
    # SDVin and SDVout are Delay history matricies in the fs=8000 domain. There
    # is one row per segment of constant delay
    # Column 0, Sample number of last sample of constant delay segment
    # Column 1, estimated Delay of segment
    # Column 2, Validity of delay estimation for segment
    # (0 for invalid, 1 for valid)
    # Convert from ms to samples
    len_t = 8 * len_t
    len_b = 8 * len_b
    len_s = 8 * len_s
    # Find number of segments of constant delay
    nsegs = np.size(SDVin, 0)
    # Find length of each segment
    seglens = np.diff(np.append(0, SDVin[:, 0]))
    # Append two to SDVin. SDVin now has
    # Column 0, Sample number of last sample of constant delay segment
    # Column 1, estimated Delay of segment
    # Column 2, Validity of delay estimation segment (0=invalid, 1=valid)
    # Column 3, Number of samples in segment
    # Column 4, Status of segment (0=needs to be considered, 1=should be
    # ingnored) Start with all zeros.
    SDVin = np.column_stack((SDVin, seglens, np.zeros(nsegs)))
    # Find location and type of the shortest segment with status 0 in SDVin
    ptr, seg_type = find_smallest_seg(SDVin)

    # If there is such a segment
    if ptr != 0:
        # Extract its length
        current_seg_len = SDVin[ptr, 3]
    else:
        # Otherwise create a fictitious segment length that is long enough
        # to prevent entering the "while loop" that follows
        current_seg_len = max([len_t, len_b, len_s]) + 1

    # While the shortest segment qualifies for consideration under at least
    # one of the three length thresholds
    while current_seg_len <= max([len_t, len_b, len_s]):
        # Find number of segments in current version of SDVin
        n = np.size(SDVin, 0)
        # If current segment is a left tail and conforms with tail threshold
        if seg_type == "LT" and current_seg_len <= len_t:
            # Join current segment to right neighbor segment
            # to create new combined segment
            SDVin = SDVin[np.setdiff1d(range(0, n), ptr), :]
            # (Setdiff is the set difference function. As called above, it
            # returns a length n-1 vector containing 1,2,...ptr-1, ptr+1, ...n)
            # Set status on the new combined segment to 0 so it receives
            # further consideration
            SDVin[ptr, 4] = False
            # Store the length of the new combined segment
            SDVin[ptr, 3] = SDVin[ptr, 3] + current_seg_len

        # If current segment is a right tail and conforms with tail threshold
        elif seg_type == "RT" and current_seg_len <= len_t:
            # Join current segment to left neighbor segment
            # to create new combined segment
            current_seg_end = SDVin[ptr, 0]
            SDVin = SDVin[np.setdiff1d(range(0, n), ptr), :]
            SDVin[ptr - 1, 0] = current_seg_end
            # Set status on the new combined segment to 0 so it receives
            # further consideration
            SDVin[ptr - 1, 4] = False
            # Store the length of the new combined segment
            SDVin[ptr - 1, 3] = SDVin[ptr - 1, 3] + current_seg_len

        # If current segment is a blip and conforms with blip threshold
        elif seg_type == "BI" and current_seg_len <= len_b:
            # Join current segment and left neighbor to right neighbor segment
            # to create new combined segment
            left_neb_len = SDVin[ptr - 1, 3]
            SDVin = SDVin[np.setdiff1d(range(0, n), [ptr - 1, ptr]), :]
            # Set status on the new combined segment to 0 so it receives
            # further consideration
            SDVin[ptr - 1, 4] = False
            # Store the length of the new combined segment
            SDVin[ptr - 1, 3] = SDVin[ptr - 1, 3] + current_seg_len + left_neb_len

        # If current segment is a step and conforms with step threshold
        elif seg_type == "SP" and current_seg_len <= len_s:
            # Join current segment to left or right neighbor or leave it
            # as is. Choice of these 3 actions depends on correlation results
            # -----------------------Preparations-------------------------------
            start = SDVin[ptr - 1, 0] + 1  # First sample of current segment
            stop = SDVin[ptr, 0]  # Last sample of current segment
            L_dly = SDVin[ptr - 1, 1]  # Delay of segment left of current segment
            C_dly = SDVin[ptr, 1]  # Delay of current segment
            R_dly = SDVin[ptr + 1, 1]  # Delay of segment right of current segment
            # Correlation at delay of left neighbor
            lcorr = single_corr(x_speech, y_speech, start, stop, L_dly)
            # Correlation at delay of right neighbor
            rcorr = single_corr(x_speech, y_speech, start, stop, R_dly)
            # Correlation at delay of current segment
            ccorr = single_corr(x_speech, y_speech, start, stop, C_dly)
            # Which of these 3 correlations is largest?
            loc = np.argmax([lcorr, rcorr, ccorr])

            # If correlation at delay of left neighbor is largest
            if loc == 0:
                # Join current segment to left neighbor segment
                # to create new combined segment
                current_seg_end = SDVin[ptr, 0]
                SDVin = SDVin[np.setdiff1d(range(0, n), ptr), :]
                SDVin[ptr - 1, 0] = current_seg_end
                # Set status on the new combined segment to 0 so it receives
                # further consideration
                SDVin[ptr - 1, 4] = False
                # Store the length of the new combined segment
                SDVin[ptr - 1, 3] = SDVin[ptr - 1, 3] + current_seg_len
            # If correlation at delay of right neighbor is largest
            elif loc == 1:
                # Join current segment to right neighbor segment
                # to create new combined segment
                SDVin = SDVin[np.setdiff1d(range(0, n), ptr), :]
                # Set status on the new combined segment to 0 so it receives
                # further consideration
                SDVin[ptr, 4] = False
                # Store the length of the new combined segment
                SDVin[ptr, 3] = SDVin[ptr, 3] + current_seg_len
            else:
                # Don't change the step, but change its status to 1 for no
                # further consideration
                SDVin[ptr, 4] = True
        # For all other segment types, this function makes no changes.
        else:
            # Change status to 1 for no further consideration
            SDVin[ptr, 4] = True
        # Find location and type of the shortest segment with status 0 in SDVin
        ptr, seg_type = find_smallest_seg(SDVin)
        # If there is such a segment
        if ptr != 0:
            # Extract its length
            current_seg_len = SDVin[ptr, 3]
        else:
            # Otherwise create a fictitious segment length that is long enough
            # to terminate the "while loop"
            current_seg_len = max([len_t, len_b, len_s]) + 1
        # End of while loop
    # SDVout contains just the first 3 columns of SDVin
    SDVout = SDVin[:, 0:3]
    # Identify the segments that reflect a change in delay or validity
    keepers = np.append(
        np.nonzero(np.logical_or(np.diff(SDVout[:, 1]) != 0, np.diff(SDVout[:, 2]) != 0)),
        np.size(SDVout, 0) - 1,
    )
    # Retain only those segments
    SDVout = SDVout[keepers, :]
    return SDVout


# ==========================================================================
def find_smallest_seg(SDVLS):
    # This function finds the smallest segment that has status 0 and reports
    # the segment type. SDVLS is a matrix with one row per segment of constant
    # delay. The 5 columns are:
    # Column 0, Sample number of last sample of constant delay segment
    # Column 1, estimated Delay of segment
    # Column 2, Validity of delay estimation segment (0=invalid, 1=valid)
    # Column 3, Length of segment in samples
    # Column 4, Status of segment (0=needs to be considered, 1=should be
    # ignored)
    #
    # ptr points to the row in SDVLS that has the smallest segment with
    # status 0. If all segments have status 0 , then ptr=0. If there is a tie
    # for shortest segment, the first segment to occur is reported
    #
    # seg_type indicates the type of segment ptr points to:
    #'IV' segment with invalid delay estimate
    #'IS' isolated, valid but neighbors on each side are invalid
    #'LT' lefthand tail, the segment to the left of an LT segment is either
    # invalid or does not exist, but segment to right is valid
    #'RT' righthand tail, the segment to the right of an RT segment is either
    # invalid or does not exist, but segment to left is valid
    #'BI' blip, segment and both neighbors are valid. Both neighbors share a
    # common delay value which is different from the delay value of the BI
    # segment
    #'SP' step, segment and both neighbors are valid and all 3 have different
    # delay values
    #'xx' is returned when ptr=0, i.e. there is no result to report.

    # Find number of segments
    nsegs = np.size(SDVLS, 0)
    # Create list of all segments with status=0
    goodlist = np.where(SDVLS[:, 4] == 0)[0]
    # If no such segment, function is done
    if goodlist.size == 0:
        ptr = 0
        seg_type = "xx"
        # There is one or more segments with status=0
    else:
        # Find the shortest such segment
        loc = np.argmin(SDVLS[goodlist, 3])
        # Set ptr accordingly
        ptr = goodlist[loc]
        # If segment is invalid, mark it as such and function is finished
        if SDVLS[ptr, 3] == 0:
            seg_type = "IV"

        # Special case for first segment
        elif ptr == 0:
            # If segment to right is valid
            if SDVLS[ptr + 1, 2] == True:
                # First segment is a left tail
                seg_type = "LT"
            else:
                # Otherwise first segment is isolated
                seg_type = "IS"

        # Special case for last segment
        elif ptr + 1 == nsegs:
            # If segment to left is valid
            if SDVLS[ptr - 1, 2] == True:
                # Last segment is a right tail
                seg_type = "RT"
            else:
                # Otherwise last segment is isolated
                seg_type = "IS"

        # All remaining segments have two neighbors
        else:
            # Check validity of segment to left of current segment
            lv = SDVLS[ptr - 1, 2]
            # Check validity of segment to right of current segment
            rv = SDVLS[ptr + 1, 2]
            # Use these two validities to identify appropriate segment type
            if lv == True and rv == False:
                seg_type = "RT"
            elif lv == False and rv == True:
                seg_type = "LT"
            elif lv == False and rv == False:
                seg_type = "IS"
            # Both neighbors are valid, so current segment is either a blip
            # or a step
            else:
                # If neighbors have the same delay, current segment is a blip
                if SDVLS[ptr - 1, 1] == SDVLS[ptr + 1, 1]:
                    seg_type = "BI"

                # Otherwise current segment is a step
                else:
                    seg_type = "SP"
    return ptr, seg_type


# ==========================================================================
def single_corr(x_speech, y_speech, start, stop, delay):
    # Usage: rho=single_corr(x_speech,y_speech,start,stop,delay)
    # This function calculates a single correlation value between
    # a segment of x_speech and y_speech. The goal is to use the
    # samples of y_speech from "start" to "stop" inclusive and the
    # corresponding portion of x_speech, but shifted forward in time
    # by the number of samples specified in "delay."
    # These segments may be shortened as necessary if not enough
    # samples are available.
    # The correlation is direct-form and is performed by the
    # function non_fft_xc_all
    # First sample of x_speech that will be used
    sstart = int(start - delay)
    lefttrim = 0
    # If that sample does not exist
    if sstart < 0:
        # It is necessary to shift the start of the correlation window
        lefttrim = -sstart
        sstart = 0
    # Last sample of x_speech that will be used
    sstop = int(stop - delay)
    ns = len(x_speech)
    righttrim = 0
    # If that sample does not exist
    if sstop > ns:
        # It is necessary to shift the end of the correlation window
        righttrim = sstop - ns
        sstop = ns
    # Extract speech from correlation window and rectify
    x = abs(x_speech[sstart:sstop])
    y = abs(y_speech[int(start + lefttrim) : int(stop - righttrim)])
    # Perform unnormalized correlation
    rho, denom = non_fft_xc_all(x, y)
    # Normalize to find correlation value
    rho = rho / denom
    return rho


# ==========================================================================
def extend_val_res(SDVin):
    # Usage: SDVout=extend_val_res(SDVin)
    # This function extrapolates valid results to cover areas where there are
    # none. SDVin and SDVout are Delay history matrices in the fs=8000 domain:
    # Column 0, Sample number of last sample of constant delay segment
    # Column 1, estimated Delay of segment
    # Column 2, Validity of delay estimation segment
    # (False for invalid, True for valid)
    #
    # For interior invalid regions, the function splits the region in half and
    # extrapolates each neighboring valid region to cover half of the invalid
    # region. For exterior invalid regions, the function extrapolates the
    # single neighboring valid region
    # Find number of segments
    nsegs = np.shape(SDVin)[0]
    # Copy input data to output data
    SDVout = SDVin
    # If there is more than one segment, then loop over all segments
    if 1 < nsegs:
        for i in range(0, nsegs):
            # If the current segment is not valid
            if not SDVout[i, 2]:
                # Leading invalid segment case
                if i == 0:
                    SDVout[0, 1] = SDVout[1, 1]
                    # Trailing invalid segment case
                elif i == nsegs - 1:
                    SDVout[-1, 1] = SDVout[-2, 1]
                    # Interior invalid segment case
                else:
                    # Half the width of the invalid segment
                    hw = round((SDVout[i, 0] - SDVout[i - 1, 0]) / 2)
                    # Extend previous segment to cover first
                    SDVout[i - 1, 0] = SDVout[i - 1, 0] + hw
                    # Copy delay of following segment to cover second half
                    SDVout[i, 1] = SDVout[i + 1, 1]
        # Find segments associated with changes in delay
        keepers = np.append(np.nonzero(np.diff(SDVout[:, 1]))[0], nsegs - 1)
        # Retain only those segments
        SDVout = np.column_stack((SDVout[keepers, 0], SDVout[keepers, 1]))
    else:
        SDVout = SDVout[0, [0, 1]]
    return SDVout


# ==========================================================================
def LSE(s, d, Df, Dv, maxsp):
    # Usage: lse_f,lse_v=LSE(s,d,Df,Dv,maxsp)
    # This function calculates log-spectra error for fixed and variable delay
    # estimates.
    #
    # s is a column vector of source speech samples (system under test input)
    # d is a column vector of distorted speech samples (system under test
    # output) s and d should have no delay compensation applied
    # Df is a fixed scalar delay value
    # Dv is a delay history matrix where each row corresponds to a segment of
    # constant delay and
    # Column 1 holds number of last sample in segment
    # Column 2 holds delay value for segment
    # Column 3 holds 1 to indicate valid delay estimate for the segment,
    # and holds 0 otherwise
    # max sp tells the max spacing between LSE computation locations in ms
    # lse_f and lse_v are the fixed and variable LSE results in dB
    #
    # If the lengths of s and d are such that delay compensation by either
    # Df or Dv leaves insufficent signal for LSE calculations, then this
    # function returns lse_f=lse_v=0.
    # Length of LSE window in samples
    lsewin = 128
    # Convert from ms to samples
    maxsp = my_round(maxsp * 8)
    # Find list of segment numbers that are valid
    goodlist = np.nonzero(Dv[:, 2])[0]
    # Find number of valid segments in Dv
    nsegs = len(goodlist)
    # Will hold center locations of LSE windows in d
    dlocs = []
    # Will hold analogous center locations of LSE windows in s, according to
    # the variable delay estimate
    slocs_var = []
    # Loop over all segments
    for i in range(0, nsegs):
        # If it is the first segment in Dv
        if goodlist[i] == 0:
            # Starting sample number must be 1
            start = 0
        else:
            # Otherwise it is 1 more than last sample of previous segment
            start = Dv[goodlist[i] - 1, 0] + 1
        # Find last sample of segment
        stop = Dv[goodlist[i], 0]
        # Find delay of the segment
        segdel = Dv[goodlist[i], 1]
        # Find center of segment
        center = my_round((stop + start) / 2)
        # Total number of LSE windows that will fit on this segment is 2*hn+1
        hn = np.floor((stop - center - 320 - lsewin / 2) / maxsp)
        # Calculate the window location(s)
        if hn >= 1:
            locs = center + (np.arange(-hn, hn + 1) * maxsp)
        else:
            locs = center
        # Append locations to the list of LSE window center locations in d
        dlocs = np.append(dlocs, locs)
        # Append locations to the list of LSE window center locations in s,
        # compensate for segment delay
        slocs_var = np.append(slocs_var, locs - segdel)
    # Create list of corresponding centers of LSE windows in s,
    # according to the fixed delay estimate
    slocs_fxd = dlocs - Df
    # Find 4 constants
    L = int(my_round(lsewin / 2))
    R = lsewin - L
    len_d = len(d)
    len_s = len(s)
    # Find locations of window centers that will result in windows that do
    # not extend beyond the ends of s or d
    goodlocs = np.nonzero(
        np.logical_and.reduce(
            (
                0 <= (dlocs - L),
                (dlocs + R) < len_d,
                0 <= (slocs_var - L),
                (slocs_var + R) < len_s,
                0 <= (slocs_fxd - L),
                (slocs_fxd + R) < len_s,
            )
        )
    )[0]
    # Retain only such locations
    dlocs = dlocs[goodlocs]
    slocs_var = slocs_var[goodlocs]
    slocs_fxd = slocs_fxd[goodlocs]
    nlocs = len(dlocs)
    # If there are locations for LSE calculations
    if 0 < nlocs:
        # Build matrices of speech samples from the desired locations
        # Each column contains speech samples for a given LSE window
        D = np.zeros((lsewin, nlocs))
        Sv = np.zeros((lsewin, nlocs))
        Sf = np.zeros((lsewin, nlocs))
        for i in range(0, nlocs):
            D[:, i] = d[np.int_(np.arange(dlocs[i] - L, dlocs[i] + R))]
            Sf[:, i] = s[np.int_(my_round(np.arange(slocs_fxd[i] - L, slocs_fxd[i] + R)))]
            Sv[:, i] = s[np.int_(np.arange(slocs_var[i] - L, slocs_var[i] + R))]
        # Generate column vector with periodic Hanning window, length is lsewin
        win = 0.5 * (1 - np.cos(2 * np.pi * np.arange(0, lsewin) / lsewin))
        win = win.reshape(len(win), 1)
        # Repeat this window in each column of the matrix Win.
        # (Win is lsewin by nlocs.)
        # Multiply by window and peform FFT on each column of each speech matrix
        D = fft(D * win, axis=0)
        Sf = fft(Sf * win, axis=0)
        Sv = fft(Sv * win, axis=0)
        # Extract magnitude of unique half of FFT result
        D = np.abs(D[range(int(np.floor(lsewin / 2)) + 1), :])
        Sf = np.abs(Sf[range(int(np.floor(lsewin / 2)) + 1), :])
        Sv = np.abs(Sv[range(int(np.floor(lsewin / 2)) + 1), :])
        # Limit results below at 1 to prevent log(0). This is below the 10 dB
        # clamping threshold used below, so these clamped samples will not be
        # used
        D = np.maximum(D, 1)
        Sf = np.maximum(Sf, 1)
        Sv = np.maximum(Sv, 1)
        # Take log and clamp results below at 10 dB. Speech peaks will
        # typically be around 100 dB, so this limits dynamic range to about 90
        # dB and prevents low level segments from inappropriately dominating the
        # LSE results
        D = np.maximum(10, 20 * np.log10(D))
        Sf = np.maximum(10, 20 * np.log10(Sf))
        Sv = np.maximum(10, 20 * np.log10(Sv))
        # Calculated LSE: inner mean is across frequency, outer mean is across
        # LSE windows (i.e. across time)
        lse_f = np.mean(np.mean(abs(D - Sf)))
        lse_v = np.mean(np.mean(abs(D - Sv)))
    # LSE calculations are not possible
    else:
        lse_f = 0
        lse_v = 0
    return lse_f, lse_v


# used in LSE
def my_round(x):
    def round_val(val):
        if val % 1 >= 0.5:
            return np.ceil(val)
        else:
            return np.floor(val)

    if isinstance(x, np.ndarray):
        for i in range(len(x)):
            x[i] = round_val(x[i])
    else:
        x = round_val(x)
    return x



#!/usr/bin/env python

import os
import sys
import platform
import json
import re
import subprocess
import argparse
import pkgutil
import shutil

# used for version checking
import pkg_resources
import mcvqoe
import mcvqoe.base

from .sync import terminal_progress_update

#name for saved settings file
settings_name = "CopySettings.json"

if platform.system() == "Windows":

    def get_drive_serial(drive):
        #args for subprocess
        sp_args={}
        #only for windows, prevent windows from appearing
        if os.name == 'nt':
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            sp_args['startupinfo'] = startupinfo
        # run vol command, seems that you need shell=True. Perhaps vol is not a real command?
        result = subprocess.run(
            f"vol {drive}", shell=True,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            **sp_args,
        )

        # check return code
        if result.returncode:
            info = result.stderr.decode("UTF-8")

            if "the device is not ready" in info.lower():
                raise RuntimeError("Device is not ready")
            else:
                raise RuntimeError(
                    f"Could not get volume info vol returnd {res.returncode}"
                )

        # find drive serial number
        m = re.search(
            "^\W*Volume Serial Number is\W*(?P<ser>(?:\w+-?)+)",
            result.stdout.decode("UTF-8"),
            re.MULTILINE,
        )

        if m:
            return m.group("ser")
        else:
            raise RuntimeError("Serial number not found")

    def list_drives():
        #args for subprocess
        sp_args={}
        #only for windows, prevent windows from appearing
        if os.name == 'nt':
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            sp_args['startupinfo'] = startupinfo

        result = subprocess.run(
            ["wmic", "logicaldisk", "get", "name"], stdout=subprocess.PIPE,
            **sp_args,
        )

        if result.returncode:
            raise RuntimeError("Unable to list drives")

        drive_table = []

        for line in result.stdout.decode("UTF-8").splitlines():
            # look for drive in line
            m = re.match("\A\s*(?P<drive>[A-Z]:)\s*$", line)
            # if there was a match
            if m:
                res = subprocess.run(
                    f'vol {m.group("drive")}',
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

                if res.returncode:
                    info = res.stderr.decode("UTF-8")

                    if "the device is not ready" in info.lower():
                        # drive is not ready, skip
                        continue
                    elif "the network path was not found" in info.lower():
                        # network drive issue, skip
                        continue
                    else:
                        raise RuntimeError(
                            f'command returnd {res.returncode} for drive '
                            f'\'{m.group("drive")}\' \'{info}\''
                        )

                # find drive label
                m_label = re.search(
                    m.group("drive").rstrip(":")
                    + "\W*(?P<sep>\w+)\W*(?P<label>.*?)\W*$",
                    res.stdout.decode("UTF-8"),
                    re.MULTILINE,
                )

                if m_label:
                    # dictionary with serial and label
                    info = {"drive": line.strip()}
                    # check if we got a label
                    if m_label.groups("sep") == "is":
                        info["label"] = m_label.groups("label")
                    else:
                        info["label"] = ""

                    m_ser = re.search(
                        "^\W*Volume Serial Number is\W*(?P<ser>(?:\w+-?)+)",
                        res.stdout.decode("UTF-8"),
                        re.MULTILINE,
                    )

                    if m_ser:
                        info["serial"] = m_ser.group("ser")
                    else:
                        info["serial"] = ""

                    drive_table.append(info)

        return tuple(drive_table)


else:

    def list_drives():
        raise RuntimeError("Only Windows is supported at this time")

    def get_drive_serial(drive):
        raise RuntimeError("Only Windows is supported at this time")


def log_update(log_in_name, log_out_name, dryRun=False, progress_update=terminal_progress_update):
    with open(log_in_name, "rt") as fin:
        # will hold extra chars from input file
        # used to allow for partial line matches
        extra = ""
        if os.path.exists(log_out_name):
            with open(log_out_name, "rt") as fout:
                for line, (lin, lout) in enumerate(zip(fin, fout), start=1):
                    # check if the last match was not a full match
                    if extra:
                        raise RuntimeError(
                            f"At line {line}, last line was a partial match."
                        )
                    # check if lines are the same
                    if lin != lout:
                        # check if lout starts with lin
                        if lin.startswith(lout):
                            # get the chars in lout but not lin
                            extra = lin[len(lout) :]
                        else:
                            raise RuntimeError(
                                f"Files '{log_out_name}' and '{log_in_name}' differ at line {line}, can not copy"
                            )

                # get the remaining data in the file
                out_dat = fout.read()
        else:
            if not dryRun:
                # make sure that path to log file exists
                os.makedirs(os.path.dirname(log_out_name), exist_ok=True)
            # no in_dat
            in_dat = None

        # get remaining data in input file
        in_dat = fin.read()

        # strip trailing white space, add extra data
        in_dat = extra + in_dat.rstrip()

        # check if we have more data from the input file
        if in_dat:

            if not dryRun:
                #copy file to new location
                shutil.copy(log_in_name,log_out_name)

            progress_update('log-complete',0 ,0, lines=len(in_dat.splitlines()), file=log_out_name)
        else:
            if out_dat:
                raise RuntimeError("Input file is shorter than output")
            else:
                progress_update('log-complete',0 ,0, lines=0, file=log_out_name)

    # print success message
    #print(f"Log updated successfully to {log_out_name}\n")

def load_settings_file(file, match_drive=True):
    with open(file, "rt") as fp_set:
        set_dict = json.load(fp_set)

    if "Direct" not in set_dict:
        # default direct to False
        set_dict["Direct"] = False

    if set_dict["Direct"]:
        set_dict['prefix'] = ""
    else:

        #if we get a string for DriveSerial, make it a tuple
        if isinstance(set_dict["DriveSerial"], str):
            set_dict["DriveSerial"]=(set_dict["DriveSerial"],)
        else:
            #turn other things (ie. list) into a tuple
            set_dict["DriveSerial"]=tuple(set_dict["DriveSerial"])

        if match_drive:
            #get a list of connected storage devices
            drives = list_drives()

            matching_drives = [item for item in drives
                                    if item["serial"] in set_dict["DriveSerial"]]

            if not matching_drives:
                raise RuntimeError('Could not find drive with serial '
                                        f'in {set_dict["DriveSerial"]}')
            elif len(matching_drives) != 1:
                raise RuntimeError(f'Found {len(matching_drives)} '
                                    'matching drives. Please unplug all but one')

            #only one drive found, get info
            drive_info = matching_drives[0]

            # create drive prefix, add slash for path concatenation
            set_dict['prefix'] = drive_info["drive"] + os.sep

    return set_dict

def create_new_settings(direct, dest_dir, cname):
    if direct:
        prefix = ""
        rel_path = dest_dir
        drive_ser = None
    else:
        # split drive from path
        (prefix, rel_path) = os.path.splitdrive(dest_dir)

        # get serial number for drive
        drive_ser = get_drive_serial(prefix)

        # add slash for path concatenation
        prefix = prefix + os.sep

    # create dictionary of options, normalize paths
    set_dict = {
        "ComputerName": os.path.normpath(cname),
        "DriveSerial": (drive_ser,),
        "Path": os.path.normpath(rel_path),
        "Direct": direct,
        "prefix": prefix,
    }

    return set_dict

def write_settings(set_dict, file):
    save_keys = ("ComputerName", "DriveSerial", "Path", "Direct")

    #filter dictionary to contain only the specified keys
    out_dict = {k : v for k,v in set_dict.items() if k in save_keys}

    #if we get a string for DriveSerial, make it a tuple
    if isinstance(out_dict["DriveSerial"], str):
        out_dict["DriveSerial"]=(out_dict["DriveSerial"],)

    #write out new dict
    json.dump(out_dict, file)

def add_drive(path, set_path):
    (prefix, rel_path) = os.path.splitdrive(path)

    if rel_path:
        #normalize rel_path to get rid of extra slashes
        rel_path = os.path.normpath(rel_path)

        if not rel_path and rel_path != os.path.sep:
            raise RuntimeError(f'Expected path to drive but got \'{path}\'')

    #load settings from file
    settings = load_settings_file(set_path, match_drive=False)

    if settings['Direct']:
        raise RuntimeError('Can not add a drive to a direct sync')

    # get serial number for the drive to add
    drive_ser = get_drive_serial(prefix)

    #get drives as a set
    drive_set = set(settings['DriveSerial'])

    #add drive serial
    drive_set.add(drive_ser)

    settings['DriveSerial'] = tuple(drive_set)

    #write new settings
    with open(set_path, 'w') as f:
        write_settings(settings, f)


def input_log_name(d):
    return os.path.join(d, "tests.log")

def output_log_name(set_dict):
    return os.path.join(
                        set_dict['prefix'],
                        set_dict["Path"],
                        set_dict["ComputerName"] + "-tests.log"
                    )

def update_sync(set_dict, sync_dir=None, dry_run=False, progress_update=terminal_progress_update):
    if sync_dir is None:
        sync_dir = os.path.join(set_dict['prefix'], "sync")

    SyncScript = os.path.join(sync_dir, "sync.py")
    sync_ver_path = os.path.join(sync_dir, "version.txt")

    sync_update = False

    if not os.path.exists(SyncScript):
        sync_update = True
        # print message
        progress_update('supdate-missing', 0, 2)

    if not sync_update:

        if os.path.exists(sync_ver_path):
            # read version from file
            with open(sync_ver_path, "r") as f:
                sync_ver = pkg_resources.parse_version(f.read())

            # get version from package
            qoe_ver = pkg_resources.parse_version(mcvqoe.base.version)

            # we need to update if sync version is older than mcvqoe version
            sync_update = qoe_ver > sync_ver

            if sync_update:
                progress_update('supdate-old', 0, 2)

        else:
            sync_update = True
            progress_update('supdate-vmissing', 0, 2)

    if sync_update and not dry_run:
        # there is no sync script
        # make sync dir
        os.makedirs(sync_dir, exist_ok=True)
        #update progress
        progress_update('supdate-write', 1, 2)
        # copy sync script
        with open(SyncScript, "wb") as f:
            f.write(pkgutil.get_data("mcvqoe.utilities", "sync.py"))

        with open(sync_ver_path, "w") as f:
            f.write(mcvqoe.base.version)
    else:
        progress_update('supdate-skip', 1, 2)

    return SyncScript

def run_drive_sync(script, out_dir, dest_dir, dry_run=False):
    # try to get path to python
    py_path = sys.executable

    if not py_path:
        # couldn't get path, try 'python' and hope for the best
        py_path = "python"

    syncCmd = [py_path, script, "--import", out_dir, dest_dir, "--cull"]

    if dry_run:
        print("Calling sync command:\n\t" + " ".join(syncCmd))
    else:
        stat = subprocess.run(syncCmd)

        if stat.returncode:
            raise RuntimeError(
                f"Failed to run sync script exit status {stat.returncode}"
            )

test_cpy_steps = {
                    'settings':'save settings',
                    'log':'copy log',
                    'supdate':'update sync',
                    'sync':'sync files',
                 }


def copy_test_files(out_dir, dest_dir=None, cname=None, sync_dir=None, dry_run=False, force=False, direct=False, progress_update=None):

    if progress_update:
        force_lib_sync = True
    else:
        force_lib_sync = False
        progress_update=terminal_progress_update

    set_file = os.path.join(out_dir, settings_name)

    log_in_name = input_log_name(out_dir)

    num_steps = len(test_cpy_steps)

    for n, (step, name) in enumerate(test_cpy_steps.items()):

        progress_update('main-update', num_steps, n, step_name=name)

        if step == 'settings':
            if os.path.exists(set_file):

                set_dict = load_settings_file(set_file)

            else:
                if not cname:
                    raise RuntimeError(
                        f"--computer-name not given and '{set_file}' does not exist"
                    )

                if not dest_dir:
                    raise RuntimeError(f"--dest-dir not given and '{set_file}' does not exist")

                # TODO : check for questionable names in path?

                set_dict = create_new_settings(direct, dest_dir, cname)

            with (
                os.fdopen(os.dup(sys.stdout.fileno()), "w")
                if dry_run
                else open(set_file, "w")
            ) as sf:
                if dry_run:
                    print("Settings file:")
                write_settings(set_dict, sf)
        elif step == 'log':
            # file name for output log file
            log_out_name = output_log_name(set_dict)

            log_update(log_in_name, log_out_name, dry_run, progress_update=progress_update)

            # create destination path
            destDir = os.path.join(set_dict['prefix'], set_dict["Path"])

        elif step == 'supdate':
            if not set_dict["Direct"]:
                #update the sync script on the drive
                sync_script = update_sync(
                                            set_dict,
                                            sync_dir=sync_dir,
                                            dry_run=dry_run,
                                            progress_update=progress_update
                                          )
        elif step == 'sync':
            if not set_dict["Direct"] and not force_lib_sync:
                run_drive_sync(sync_script, out_dir, destDir, dry_run=dry_run)
            else:
                # direct sync, use library version
                from .sync import import_sync

                if dry_run:
                    print(
                        "Calling sync command:\n\t"
                        + f"sync.sync_files({repr(out_dir)}, {repr(destDir)}, bd=False, cull=True, sunset=30)"
                    )
                else:
                    import_sync(out_dir, destDir, bd=False, cull=True, sunset=30, progress_update=progress_update)

def recursive_sync(out_dir, dry_run=False, sync_dir=None, progress_update=None):
    #keep track of how many directories we found
    num_found = 0
    num_success = 0
    #check if progress update was given
    if progress_update is None:
        #use terminal function by default
        prog_fun=terminal_progress_update
    else:
        #use given function
        prog_fun=progress_update
    #get directory path
    out_dir = os.path.abspath(out_dir)
    for n, (root, dirs, files) in enumerate(os.walk(out_dir, topdown=True)):
        if dry_run:
            print(f'Checking "{root}" for "{settings_name}"')
        #check for copy settings
        if settings_name in files:
            num_found += 1
            prog_fun('recur-found', 0, n, dir=root)
            try:
                #settings found, copy files
                copy_test_files(root,dry_run=dry_run, sync_dir=sync_dir, progress_update=progress_update)
                #no error, this was a success
                num_success += 1
            except RuntimeError as e:
                #print error and continue
                prog_fun('recur-error', 0, n, err=str(e))
            #remove directories from dirs
            #this will skip all directories
            dirs.clear()
    #return stats
    return num_found, num_success

# main function
def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-d",
        "--dest-dir",
        default=None,
        type=str,
        metavar="DIR",
        dest="dest_dir",
        help="Path to store files on removable drive",
    )
    parser.add_argument(
        "-o",
        "--outdir",
        default="",
        metavar="DIR",
        help="Directory where test output data is stored",
    )
    parser.add_argument(
        "-c",
        "--computer-name",
        default=None,
        metavar="CNAME",
        dest="cname",
        help="computer name for log file renaming",
    )
    parser.add_argument(
        "-s",
        "--sync-directory",
        default=None,
        metavar="SZ",
        dest="sync_dir",
        help="Directory on drive where sync script is stored",
    )
    parser.add_argument(
        "-D",
        "--dry-run",
        action="store_true",
        default=False,
        dest="dry_run",
        help="Go through all the motions but, don't copy any files",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="overwrite config files with values from arguments",
    )
    parser.add_argument(
        "-i",
        "--direct",
        action="store_true",
        default=False,
        help="Copy directly to destination, not to a HD",
    )
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        default=False,
        help="Find copy settings recursively and sync where they are found",
    )

    # parse arguments
    args = parser.parse_args()

    if args.outdir:
        out_dir = os.getcwd()
    else:
        out_dir = args.outdir


    #convert to dict for use as kwargs
    args_dict = vars(args).copy()

    #remove some things
    args_dict.pop('outdir')
    args_dict.pop('recursive')

    if args.recursive:
        num_found, num_success = recursive_sync(out_dir, dry_run=args.dry_run, sync_dir=args.sync_dir)
        #check if we found any files
        if num_found:
            print(f'{num_found} test directories found, {num_success} successfully synced')
        else:
            print('No test directories found. Has sync been set up? are you in the correct directory?')
    else:
        copy_test_files(out_dir,**args_dict)


if __name__ == "__main__":
    main()



from .timecode import timecode_types


def require_timecode(audio_player):
    """
    Throw error if a timecode will not be recorded.

    This function is used to make sure that a timecode will be recorded for
    a two location test.
    
    Parameters
    ----------
    audio_player: AudioPlayer or QoEsim class
        object to check recording channels on.
    
    Raises
    ------
    ValueError
        If there are no timecode channels

    See Also
    --------
    mcvqoe.mouth2ear : Two location m2e tests use timecodes.

    Examples
    --------

    By default there is no timecode recorded by AudioPlayer.

    >>> ap=AudioPlayer()
    >>> ap.require_timecode()
    ValueError: Timecode channel not found in ('rx_voice',)

    If a timecode is recorded, no error is raised.
    >>> ap=AudioPlayer(rec_chans={'rx_voice':0,'IRIGB_timecode':1})
    >>> ap.require_timecode()

    """
    
    for chan in audio_player.rec_chans:
        if 'timecode' in chan:
            #found! we are done
            break
    else:
        #not found, raise error
        chans=tuple(audio_player.rec_chans.keys())
        raise ValueError(f'Timecode channel not found in {chans}')

def timecode_chans(chans,tc_priority=tuple(timecode_types.keys())):
    '''
    Return the indices of the timecode recording channels.
    
    Finds the channels that contain timecode. Sorts timecodes based on
    tc_priority.
    
    Parameters
    ----------
    chans : tuple of strings
        Recording channels to search.
    tc_priority : tuple of strings, default=('IRIGB_timecode','soft_timecode')
        Matching timecodes in this array get sorted first based on order in array.
    '''
    
    tc=[]
    order=[]
    
    for n,chan in enumerate(chans):
        if('timecode' in chan):
            #found timecode, append to array
            tc.append(n)
            try:
                p=tc_priority.index(chan)
            except ValueError:
                #not found, lowest priority
                p=len(tc_priority)
            order.append(p)
            
    #TODO : sort by order
    return tc


import os
import fnmatch
import mcvqoe.base
import statistics
import argparse
import numpy as np
import plotly.graph_objects as go
import pandas as pd
import plotly.io as pio
pio.renderers.default = 'browser'
import plotly.express as px
import math
import re
import json

def Test_Stats(Wav_Dir):
    """
    Test_Stats Read in a directory of WAV files. 
    Use a-weight, FSF, and clipping checks to inform
    user of potential problems in collected data. Flags
    trials that may require further investigation.
    
    Parameters
    ----------
    Wav_Dir : string
        directory of WAV files
        
    Returns
    -------
    TX_filename : list
        names of tx audio files
    tx_wavs   :  list
        audio for tx audio
    Trials : int
        number of trials 
    rx_rec : list 
        audio for rx recordings   
    fs : int 
        sampling rate of rx recordings       
    rx_dat : list
        names of rx recordings   
    bad_trial : set
        Names of all trials that set off any of the flags for 
        potential problems
    """
    # Get trial recordings
    print("Loading recordings")
    # Get all the Rx wav files 
    Dir_Files = os.listdir(Wav_Dir)
    Naming = 'Rx*'
    all_wavs = fnmatch.filter(Dir_Files, Naming)
    # Get total number of trials
    Trials = len(all_wavs)
    # Inform user of number of data sets
    print("Number of trials:")
    print(Trials)
    # Create empty list for rx recordings 
    rx_rec = []
    rx_dat = []
    # Cycle through, in order
    for n in range(1,Trials+1):
        start = 'Rx'+str(n)+'_'
        rx_name = [s for s in all_wavs if start in s]
        rx_path = Wav_Dir + '/' + rx_name[0]
        fs,y_rec = mcvqoe.base.audio_read(rx_path)
        rx_rec.append(y_rec[:]) 
        rx_dat.append(rx_name[0])
    # Create bad trial set
    bad_trial = set()
    # Find all the Tx files in the wav_dir, strip 
    # the Tx and .wav off 
    TX_names = 'Tx*'
    TX_obj = fnmatch.filter(Dir_Files, TX_names)
    TX_filename =fnmatch.filter(TX_obj, '*.wav')
    # Create empty list for tx wavs 
    tx_wavs = []
    # Cycle through and get all the TX files
    for k in range(0,len(TX_filename)):
       tx_path = Wav_Dir + '/' + TX_filename[k]
       fs,tx_wavfile = mcvqoe.base.audio_read(tx_path)
       tx_wavs.append(tx_wavfile)
   
    return TX_filename, tx_wavs, Trials, rx_rec, fs, rx_dat, bad_trial
    
def AW_Rec_Check(Trials, rx_rec, fs, rx_dat, bad_trial):
    """
    Calculates the a-weight (dBA) of each trial. Plot these  
    values. Check for trials with a dBA less than - 60 dBA.
    Check for trials with a dBA a certain distance from the mean.
    
    Parameters
    ----------
    Trials : int
        number of trials 
    rx_rec : list 
        audio for rx recordings   
    fs : int 
        sampling rate of rx recordings       
    rx_dat : list
        names of rx recordings   
    bad_trial : set
        Names of all trials that set off any of the flags for 
        potential problems

    Returns
    -------
    A_Weight : array
        A_Weight of every trial 
    bad_trial : set
        Names of all trials that set off any of the flags for 
        potential problems      
    AW_flag : set
        Trials that have low dBA values (and likely lost audio)
        or otherwise deviate from the patterns of the dataset 
    """
    # Create empty list, set for a-weight values and flag     
    A_Weight = []
    AW_flag = set()
      
    # Get A-weight
    print("Gathering Statistics and Creating Plots") 
    for k in range(0,Trials): 
        # Calculate the A-weighted power of each recording 
        aw = mcvqoe.base.a_weighted_power(rx_rec[k], fs) 
        A_Weight.append(aw)

    # Plot a-weighted power for all trials  
    A_Weight = np.asarray(A_Weight)  
    x_axis = list(range(1,Trials+1))
    dfAW = pd.DataFrame({"A-Weight": A_Weight,
                         "Trials": x_axis})  
    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x = dfAW['Trials'], 
            y = dfAW['A-Weight'],
            mode = 'markers'
        )
    )    
    fig.update_layout(title_text='A-Weighted Power of Received Audio')
    fig.update_xaxes(title_text='Trial Number')
    fig.update_yaxes(title_text='A-Weight (dBA)')
    fig.show() 

    # Calculate mean a-weight, standard deviation.
    # Use this info to find trials that may have lost
    # audio. 
    AW_lin = 10**(A_Weight/20)
    AW_low = 10**(-60/20)
    AW_Mean = round(statistics.mean(AW_lin),3)
    AW_std = round(statistics.stdev(AW_lin),3)
    for m in range(0,Trials):
        # Cycle through AW values, look for trials where the values
        # stand out by being a certain distance from the mean
        if abs(AW_lin[m]-AW_Mean) > 2*AW_std:
            # Get the name of the wav file
            AWflag_wav = rx_dat[m]
            # Add it to the bad list
            bad_trial.add(AWflag_wav)
            AW_flag.add(AWflag_wav)
        # Add a flag if the a-weight is below -60 dBA
        if AW_lin[m] < AW_low:
           # Get the name of the wav file
           AWflag_wav = rx_dat[m]
           # Add it to the bad list
           AW_flag.add(AWflag_wav)

    return A_Weight, AW_flag, bad_trial

def FSF_Rec_Check(TX_filename, tx_wavs, Trials, rx_rec, fs, rx_dat, bad_trial):   
    """
    Calculate FSF scores, standard deviation. Use this info to find trials that may have lost
    audio.
    
    Parameters
    ----------
    TX_filename : list
        names of tx audio files
    tx_wavs   :  list
        audio for tx audio
    Trials : int
        number of trials 
    rx_rec : list 
        audio for rx recordings   
    fs : int 
        sampling rate of rx recordings       
    rx_dat : list
        names of rx recordings   
    bad_trial : set
        Names of all trials that set off any of the flags for 
        potential problems

    Returns
    -------
    bad_trial : set
        Names of all trials that set off any of the flags for 
        potential problems
    FSF_all : list
       FSF scores of every trial 
    FSF_flag : set
        Trials that have low FSF scores or otherwise deviate 
        from the patterns of the dataset     
    """
    # Create empty list,set for FSF scores and flag    
    FSF_all = []
    FSF_flag = set()
    
    # Get FSF scores, plot trials
    print("Gathering FSF data") 
    # Cycle through, match RX names to tx names
    # Get just the names of tx files, remove the Tx and .wav 
    tx_base = []
    for n in range(0,len(TX_filename)):
        tx_justname = re.sub('\.wav$', '', TX_filename[n])
        tx_justname = tx_justname[3:]
        tx_base.append(tx_justname)   

    for j in range(0,Trials): 
        # Find RX files with the matching tx name, create groups 
        match_wavs = re.match(r'(Rx\d+_(?P<tx_base_name>[^.]+))',rx_dat[j])
        # find the index of the TX and RX clips to match with the lists of 
        # wav data
        TX_idx = tx_base.index(match_wavs.group('tx_base_name'))
        TX_wav = tx_wavs[TX_idx]
        RX_wav = rx_rec[j]
        # Get FSF scores for each tx-rx pair
        get_fsf = mcvqoe.base.fsf(RX_wav,TX_wav,fs)
        # Get just the FSF score
        FSF_all.append(get_fsf[0])   
        
    # Gather metrics for FSF scores
    FSF_Mean = round(statistics.mean(FSF_all),3)
    FSF_std = round(statistics.stdev(FSF_all),3)
    for m in range(0,Trials):
        # Cycle through FSF scores, look for trials where the score
        # stands out by being a certain distance from the mean
        if abs(FSF_all[m]-FSF_Mean) > 2*FSF_std:
            # Get the name of the wav file
            FSFflag_wav = rx_dat[m]
            # Add it to the bad list
            bad_trial.add(FSFflag_wav)
            FSF_flag.add(FSFflag_wav)
        # Add a flag for low FSF scores that may be a sign of 
        # dropped audio
        if FSF_all[m] < 0.3:
            # Add it to the bad list
            bad_trial.add(FSFflag_wav)
            FSF_flag.add(FSFflag_wav)
            
    # Plot FSF values 
    FSF_Scores = np.asarray(FSF_all)  
    x_axis = list(range(1,Trials+1))
    dfFSF = pd.DataFrame({"FSF Score": FSF_Scores,
                          "Trial": x_axis})  
    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
        x = dfFSF['Trial'],    
        y = dfFSF['FSF Score'],
        mode = 'markers'
        )
    ) 
    fig.update_layout(title_text='FSF Score of Received Audio')
    fig.update_xaxes(title_text='Trial Number')
    fig.update_yaxes(title_text='FSF Score')
    fig.show()

    return bad_trial, FSF_all, FSF_flag

def Clip_Rec_Check(Trials, rx_rec, rx_dat, bad_trial):
    """
    Cycle through all rx recordings and check if any clipped.
    Parameters
    ----------
    Trials : int
        number of trials 
    rx_rec : list 
        audio for rx recordings         
    rx_dat : list
        names of rx recordings   
    bad_trial : set
        Names of all trials that set off any of the flags for 
        potential problems

    Returns
    ------- 
    peak_dbfs : list
        peak amplitude of each trial, dB relative to full 
        scale 
    Clip_flag : set
        Trials that clipped 
    bad_trial : set
        Names of all trials that set off any of the flags for 
        potential problems
    """
    # Set up warning threshold
    vol_high = -1
    # Create empty list, set for peak volume and flag    
    peak_dbfs = []
    Clip_flag = set()
    for n in range(0,Trials):
        # Check for positive and negative clipping
        peak = max(abs(rx_rec[n]))
        peak_db = round(20 * math.log10(peak), 2)
        peak_dbfs.append(peak_db)
        # If approaching clipping, flag as a bad trial
        if peak_dbfs[n] > vol_high :
            # Get the name of the wav file
            clip_wav = rx_dat[n]
            # Add it to the bad list
            bad_trial.add(clip_wav)
            Clip_flag.add(clip_wav)
    
    return peak_dbfs, Clip_flag, bad_trial        
    
def Wav_Plot(rx_rec, Trials, fs):
    """
    Plot rx trial audio recordings
    
    Parameters
    ----------
    Trials : int
        number of trials 
    rx_rec : list 
        audio for rx recordings   
    fs : int 
        sampling rate of rx recordings       

    Returns
    -------
 
    """
    # Create empty list for time (seconds)
    tsec = []
    # Prepare time info for plotting
    for k in range(0,Trials):
        t = len(rx_rec[0])/fs
        ts = np.arange(0,t,1/fs)
        tsec.append(ts)
        # Plot recording
        dfWavs= pd.DataFrame({"time":ts, 
                        "audio":rx_rec[0]})
    dfWavs.set_index('time')
    fig = px.line(dfWavs, y='audio',x='time')
    fig.update_layout(title_text='Trial Recordings')
    fig.update_xaxes(title_text='Time (s)')
    fig.update_yaxes(title_text='Amplitude')
    fig.show()  
    
 
def Problem_Trials(bad_trial, Clip_flag, AW_flag, FSF_flag):     
    """
    Warn user of problem trial recording names 
    
    Parameters
    ----------
    bad_trial : set
        Names of all trials that set off any of the flags for 
        potential problems
    Clip_flag : set    
        names of trials flagged for clipping
    AW_flag : set
        names of trials flagged for their dBA values
    FSF_flag : set     
        names of trials flagged for their FSF scores   

    Returns
    -------

    """
    np.disp('The following trials were flagged as potentially having issues')
    np.disp(bad_trial)
    if Clip_flag:
        np.disp('The following trials were flagged for clipping')
        np.disp(Clip_flag)  
    if AW_flag:    
        np.disp('The following trials were flagged for their a-weight')
        np.disp(AW_flag)
    if FSF_flag:    
        np.disp('The following trials were flagged for their FSF scores')
        np.disp(FSF_flag) 

def Gather_Diagnostics(rx_dat,A_Weight,FSF_all,peak_dbfs):
    """
    Create a dataframe of all diagnostic data. A-weight,
    FSF scores, max clip amplitude. Convert to json
    
    Parameters
    ----------
    rx_dat : list
        Names of all RX trials
    A_Weight : list    
        A-Weight across all RX trials 
    FSF_all : list
        FSF scores across all RX trials
    peak_dbfs : list     
        Peak amplitude across all RX trials   

    Returns
    -------
    diagnostics_dat : json 
        Dataframe containing all the dat for diagnostics
   """
   # Create dataframe of info
    df_Diagnostics = pd.DataFrame({"RX_Name":rx_dat, 
                    "A_Weight":A_Weight,
                    "FSF_Scores":FSF_all,
                    "Amplitude":peak_dbfs})
    # Creat json file
    diagnostics_json = df_Diagnostics.to_json()
    Test_Dir = 'C:/Users/cjg2/Documents/MCV'
    with open(os.path.join(Test_Dir,'diagnostics_json.json'),'w') as f:
        f.write(diagnostics_json)      
    
    return diagnostics_json    

def main():   
    """
    Read in a directory of WAV files. 
    Use a-weight, FSF, and clipping checks to inform
    user of potential problems in collected data. Flags
    trials that may require further investigation.

    Parameters
    ----------
    Wav_Dir : string
        directory of WAV files

    Returns
    -------

    """
    # Input parsing
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-wd",
        "--Wav_Dir",
        default=None,
        type=str,
        dest="Wav_Dir",
        help="Directory to test data wav files",
    )
    # Parse input arguments
    args = parser.parse_args()

if __name__ == "__main__":
    main()
  